//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains functions that relate to the pricing of constant
// maturity swap (CMS) derivatives.
//============================================================================
//============================================================================
#include <windows.h>

#include "XllAddIn.h"

//=============================================================
int bin_name(char *name, int create, void *data, long len)
{
	if(!name) 
		return 0;

	cpp_xloper Name(name), RetVal;

	if(create)
	{
		cpp_xloper Big(data, len);
		if(RetVal.Excel(xlDefineBinaryName, 2, &Name, &Big) != xlretSuccess)
			return 0;
	}
	else
	{
		RetVal.Excel(xlDefineBinaryName, 1, &Name);
	}
	return 1;
}
//=============================================================
int __stdcall bin_exists(char *name)
{
#pragma EXPORT

	if(!name)
		return 0;

	cpp_xloper Name(name);
	cpp_xloper Big;
	return Big.Excel(xlGetBinaryName, 1, &Name) && Big.IsBigData();
}
//=============================================================
int get_binary_data(char *name, void * &data, long &len)
{
	if(!name)
		return 0;

	cpp_xloper Name(name);
	cpp_xloper Big;

	if(Big.Excel(xlGetBinaryName, 1, &Name) != xlretSuccess
	|| !Big.IsBigData())
		return 0;

	if(gExcelVersion12plus)
	{
		xloper12 *p_big = Big.OpAddr12();
		len = p_big->val.bigdata.cbData;
		if(!(data = malloc(len)))
			return 0;

		void *p = GlobalLock(p_big->val.bigdata.h.hdata);
		memcpy(data, p, len);
		GlobalUnlock(p_big->val.bigdata.h.hdata);
	}
	else
	{
		xloper *p_big = Big.OpAddr();
		len = p_big->val.bigdata.cbData;
		if(!(data = malloc(len)))
			return 0;

		void *p = GlobalLock(p_big->val.bigdata.h.hdata);
		memcpy(data, p, len);
		GlobalUnlock(p_big->val.bigdata.h.hdata);
	}
	((unsigned char *)data)[len-1] = 0; // work-around for possible bug that corrupts the null originally saved
	return 1;
}
//=============================================================
xloper * __stdcall set_bin_string(char *name, xloper *p_string)
{
#pragma EXPORT

	int create = (p_string->xltype == xltypeStr ? 1 : 0);

	if(create)
	{
		long len = (BYTE)p_string->val.str[0] + 1; // Include null
		char *p = p_string->val.str + 1; // Start of string

		if(bin_name(name, create, p, len))
			return p_xlTrue;

		return p_xlErrValue; // couldn't create
	}

	if(bin_name(name, 0, NULL, 0))
		return p_xlErrName; // deleted ok
	else
		return p_xlErrValue; // couldn't delete
}
//=============================================================
xloper * __stdcall get_bin_string(char *name)
{
#pragma EXPORT

	void *string;
	long len;

	if(get_binary_data(name, string, len))
	{
// Constructor will truncate if too long
		cpp_xloper RetVal((char *)string);
		return RetVal.ExtractXloper();
	}
	return p_xlErrName;
}