//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains examples of XLL interface functions.
//============================================================================
//============================================================================
#include <windows.h>

#include "XllAddIn.h"

#define MAX_ARG1			100.0
#define MIN_RTN_VALUE		0.00000001
#define MAX_RTN_VALUE		100.0

bool __stdcall example_core_fn(double arg_1, bool arg_2, double &rtn_value)
{
#pragma EXPORT
	if(arg_2)
		rtn_value = arg_1;
	else
		rtn_value = arg_1 * arg_1;
	return true;
}

xloper * __stdcall ExampleXllFn(double Arg1, xloper *pArg2)
{
#pragma EXPORT

// Step 1:
// If required, check to see if called from an Excel dialog, or
// check that the required add-in resources have been properly
// initialised.
	if(called_from_Excel_dlg())
		return p_xlErrValue; // in this example, fail with #VALUE!

// Step 2:
// Convert any arguments from input types to types expected by
// core function and check that required inputs exist and are
// within the correct limits.
	if(Arg1 <= 0.0 || Arg1 > MAX_ARG1)
		return p_xlErrNum;

// Declare a variable for the second core function argument.
// Assume here that false is the default value
	bool core_boolean_arg = false;

	switch(pArg2->xltype)
	{
	case xltypeNum:
		if(pArg2->val.num != 0.0)
			core_boolean_arg = true;
		break;

	case xltypeBool:
		if(pArg2->val.xbool != 0)
			core_boolean_arg = true;
		break;

	case xltypeStr:
		if((BYTE)pArg2->val.str[0] > 0
		&& (pArg2->val.str[1] == 'T' || pArg2->val.str[1] == 't'))
			core_boolean_arg = true;
		break;

	default: // Not a type that this function can convert
		return p_xlErrNum;
	}

// Step 2:
// Convert any arguments from input units to units expected by
// core function.
	Arg1 /= 100.0;

// Step 3:
// Call the core function and check the return value.

// Declare a variable for the core function return value.
	double rtn_value;

	if(example_core_fn(Arg1, core_boolean_arg, rtn_value) == false
	|| rtn_value < MIN_RTN_VALUE || rtn_value > MAX_RTN_VALUE)
		return p_xlErrValue;

// Step 4:
// Convert the return value to units promised by Excel add-in funtion
	rtn_value *= 100.0;

// Step 5:
// Convert the return value back to the right data type for this
// model, i.e., an xloper

	cpp_xloper RetVal(rtn_value);
	return RetVal.ExtractXloper();
}
//=================================================================
// Excel worksheet cell error codes are passed via VB OLE Variant
// arguments in 'ulVal'. These are equivalent to the offset below
// plus the value defined in "xlcall.h"
//=================================================================

// This is easier than using the VT_SCODE variant property 'scode'.
#define VT_XL_ERR_OFFSET		2148141008ul

VARIANT __stdcall ExampleVtFn(double Arg1, VARIANT *pArg2)
{
#pragma EXPORT

	VARIANT return_vt;
	VariantInit(&return_vt); // type is set to VT_EMPTY

// Step 1:
// If required, check to see if called from an Excel dialog, or
// check that the required add-in resources have been properly
// initialised.

// Step 2:
// Convert any arguments from input types to types expected by
// core function and check that required inputs exist and are
// within the correct limits.
	if(Arg1 <= 0.0 || Arg1 > MAX_ARG1)
	{
		return_vt.vt = VT_ERROR;
		return_vt.ulVal = VT_XL_ERR_OFFSET + xlerrNum;
		return return_vt;
	}

// Declare a variable for the second core function argument.
// Assume here that false is the default value
	bool core_boolean_arg = false;

	if(pArg2 == NULL)
	{
		return_vt.vt = VT_ERROR;
		return_vt.ulVal = VT_XL_ERR_OFFSET + xlerrNum;
		return return_vt;
	}

	switch(pArg2->vt)
	{
	case VT_R8:
		if(pArg2->dblVal != 0.0)
			core_boolean_arg = true;
		break;

	case VT_BOOL:
		if(pArg2->boolVal != 0)
			core_boolean_arg = true;
		break;

// NOTE: Called from Excel/VBA so string is byte-string
// not wide-char string, so need C cast to (char *) to
// read it.
	case VT_BSTR:
		{
			char *text = (char *)pArg2->bstrVal;

			if(text != NULL && (((BYTE)text[0] | 0x20) == 't'))
				core_boolean_arg = true;
			break;
		}

	default: // Not a type that this function can convert
		return_vt.vt = VT_ERROR;
		return_vt.ulVal = VT_XL_ERR_OFFSET + xlerrNum;
		return return_vt;
	}

// Step 3:
// Convert any arguments from input units to units expected by
// core function.
	Arg1 /= 100.0;

// Step 4:
// Call the core function and check the return value.

// Declare a variable for the core function return value.
	double rtn_value;

	if(example_core_fn(Arg1, core_boolean_arg, rtn_value) == false
	|| rtn_value < MIN_RTN_VALUE || rtn_value > MAX_RTN_VALUE)
	{
		return_vt.vt = VT_ERROR;
		return_vt.ulVal = VT_XL_ERR_OFFSET + xlerrValue;
		return return_vt;
	}

// Step 5:
// Convert the return value to units promised by Excel add-in funtion
	rtn_value *= 100.0;

// Step 6:
// Convert the return value back to the right data type for this
// model, i.e., an xloper

	return_vt.vt = VT_R8;
	return_vt.dblVal = rtn_value;
	return return_vt;
}
//=================================================================
// This example presents a cpp_xloper-wrapped interface to the real
// core function, so that the various steps are coded only once, and
// can be called by a VBA/Variant export, or an XLL/xloper or
// XLL/xloper12 export.
bool example_xll_core_fn(cpp_xloper &RetVal, double Arg1, cpp_xloper &Arg2)
{
// Step 1:
// If required, check to see if called from an Excel dialog, or
// check that the required add-in resources have been properly
// initialised.
	if(called_from_Excel_dlg())
	{
		RetVal.SetToError(xlerrValue);
		return false; // in this example, fail with #VALUE!
	}
		
// Step 2:
// Convert any arguments from input types to types expected by
// core function and check that required inputs exist and are
// within the correct limits.
	if(Arg1 <= 0.0 || Arg1 > MAX_ARG1)
	{
		RetVal.SetToError(xlerrNum);
		return false; // in this example, fail with #VALUE!
	}

// Declare a variable for the second core function argument.
// Assume here that false is the default value
	bool core_boolean_arg = false;

	if(Arg2.IsNum())
	{
		if((double)Arg2 != 0.0)
			core_boolean_arg = true;
	}
	else if(Arg2.IsBool())
	{
		core_boolean_arg = (bool)Arg2;
	}
	else if(Arg2.IsStr())
	{
		if(Arg2.First() == L'T' || Arg2.First() == L't')
			core_boolean_arg = true;
	}
	else
	{
// Not a type that this function can convert
		RetVal.SetToError(xlerrNum);
		return false;
	}

// Step 3:
// Convert any arguments from input units to units expected by
// core function.
	Arg1 /= 100.0;

// Step 4:
// Call the core function and check the return value.

// Declare a variable for the core function return value.
	double rtn_value;

	if(example_core_fn(Arg1, core_boolean_arg, rtn_value) == false
	|| rtn_value < MIN_RTN_VALUE || rtn_value > MAX_RTN_VALUE)
	{
		RetVal.SetToError(xlerrValue);
		return false;
	}

// Step 5:
// Convert the return value to units promised by Excel add-in funtion
	rtn_value *= 100.0;

// Step 6:
// Convert the return value back to the right data type for this
// model, i.e., an xloper
	RetVal = rtn_value;
	return true;
}
//=================================================================
// Wraps bool example_core_fn() and hides the core function's
// data types using VARIANTS
//=================================================================
VARIANT ExampleWrapper(VARIANT Arg1, VARIANT Arg2, VARIANT *pRtnVal)
{
	VARIANT return_vt;
	VariantInit(&return_vt); // type is set to VT_EMPTY
	return_vt.vt = VT_BOOL;

// Mimimal type checking only done here
	if(Arg1.vt != VT_R8 || Arg2.vt != VT_BOOL)
	{
		return_vt.boolVal = 0; // False
		return return_vt;
	}

// Call the core function and return the return value.
// Do not check the return values - this is left the
// caller of this wrapper function.

// Declare a variable for the core function return value.
	double rtn_value;
	bool b;

	b = example_core_fn(Arg1.dblVal, Arg2.boolVal==1, rtn_value);

// Convert true to 1, false to 0
	return_vt.boolVal = (b ? 1 : 0);

	VariantInit(pRtnVal); // type is set to VT_EMPTY
	pRtnVal->vt = VT_R8;
	pRtnVal->dblVal = rtn_value;

	return return_vt;
}
