//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
#pragma once

#include <windows.h>

// Simple thread-safe FILO stack

struct simple_stack
{
	simple_stack(int max_size)
	{
		InitializeCriticalSection(&cs_stack);
// Need to enter the CS here in case the constructor is explicitly invoked
		EnterCriticalSection(&cs_stack);
		if(stack)
		{
			free(stack);
			stack = NULL;
			size = index = 0;
		}

		if(max_size)
		{
			stack = (double *)calloc(max_size, sizeof(double));
			size = max_size;
		}
		LeaveCriticalSection(&cs_stack);
	}
	~simple_stack(void)
	{
// Need to enter the CS here in case the destructor is explicitly invoked
		EnterCriticalSection(&cs_stack);
		if(stack)
		{
			free(stack);
			stack = NULL;
			size = index = 0;
		}
		LeaveCriticalSection(&cs_stack);
		DeleteCriticalSection(&cs_stack);
	}
	bool push(double d)
	{
		EnterCriticalSection(&cs_stack);
		if(stack && index < size)
		{
			stack[index++] = d;
			LeaveCriticalSection(&cs_stack);
			return true;
		}
		LeaveCriticalSection(&cs_stack);
		return false;
	}
	bool pop(double &d)
	{
		EnterCriticalSection(&cs_stack);
		if(stack && index > 0)
		{
			d = stack[--index];
			LeaveCriticalSection(&cs_stack);
			return true;
		}
		LeaveCriticalSection(&cs_stack);
		return false;
	}
private:
	CRITICAL_SECTION cs_stack;
	double *stack;
	int index;
	int size;
};

