//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains a simple Black option class.
//============================================================================
//============================================================================
#include <windows.h>
#include <math.h>

#include "XllAddIn.h"
#include "Black.h"

#define ROOT_2PI		(2.506628274631)
#define n(x) (exp(-0.5 * (x) * (x)) / ROOT_2PI)
double __stdcall ndist(double d);

bool BlackOption::Calc(bool calc_derivs)
{
	if(m_Strike <= 0.0 || m_Fwd <= 0.0 || m_Texp < 0.0)
		return false;

	if(m_Texp == 0.0) // Intrinsic value only
	{
		if(m_Intrinsic > 0.0) // put in-the-money
		{
			m_Put = m_Intrinsic;
			m_PutDelta = 1.0;
			m_Call = m_CallDelta = 0.0;
		}
		else if(m_Intrinsic < 0.0)
		{
			m_Put = m_PutDelta = 0.0;
			m_Call = -m_Intrinsic;
			m_CallDelta = 1.0;
		}
		else
		{
			m_Put = m_PutDelta = m_Call = m_CallDelta = 0.0;
		}
		m_Straddle = m_Call + m_Put;
		m_Gamma = MAX_DOUBLE;  // really infinity but...
		m_Vega = 0.0;
		return true;
	}

	if(m_Vol <= 0.0)
		return false;

	double d1 = log(m_Fwd / m_Strike) / m_vRootT + m_vRootT / 2.0;
	double N1 = ndist(d1); // ndist(x) = N(x)
	double N2 = ndist(d1 - m_vRootT);

	m_Call = m_Fwd * N1 - m_Strike * N2;
	m_Put = m_Call + m_Intrinsic;
	m_Straddle = m_Call + m_Put;

	if(!calc_derivs)
		return true;

// Hedge calculations assume constant volatility
	double n1 = n(d1);  // n = dN(x)/dx
	m_PutDelta = (m_CallDelta = N1) - 1.0;
	m_Gamma = n1 / m_vRootT / m_Fwd / m_Vol;
	m_Vega = n1 * m_vRootT * m_Fwd;
	return true;
}
//=================================================
// XLL-exportable interface function
//=================================================
xloper * __stdcall BlackOpt(double t_exp, double fwd, double strike, double vol)
{
#pragma EXPORT

	BlackOption Opt(t_exp, fwd, strike, vol);
	Opt.Calc(true); // true: calc greeks

#define NUM_BLACK_RTN_VALS	6

	double ret_vals[NUM_BLACK_RTN_VALS] = {
		Opt.GetCallPrice(), Opt.GetPutPrice(),
		Opt.GetCallDelta(), Opt.GetPutDelta(),
		Opt.GetGamma(), Opt.GetVega()};

// Return a single row array
	cpp_xloper RetVal((RW)1, (COL)NUM_BLACK_RTN_VALS, ret_vals);
	return RetVal.ExtractXloper();
}
