//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
#pragma once

class BlackOption
{
public:
	BlackOption(double t_exp, double fwd, double strike, double vol)
		: m_Texp(t_exp), m_Fwd(fwd), m_Strike(strike), m_Vol(vol)
	{
		if(m_Texp < 0.0)
			m_Texp = 0.0;

		m_Intrinsic = m_Strike - m_Fwd; // intrinsic value of put
		m_vRootT = m_Vol * (m_RootT = sqrt(m_Texp));
	}

	BlackOption() {memset(this, 0, sizeof(BlackOption));}

	void SetVol(double vol) {m_vRootT = (m_Vol = vol) * m_RootT;}
	void SetFwd(double fwd) {m_Intrinsic = m_Strike - (m_Fwd = fwd);}
	void SetStrike(double strike) {m_Intrinsic = (m_Strike=strike)-m_Fwd;}
	void SetTexp(double t_exp)
	{
		if(t_exp < 0.0)
			return;
		m_vRootT = m_Vol * (m_RootT = sqrt(m_Texp = t_exp));
	}
	bool Calc(bool calc_derivs = false);
	double GetCallPrice()	{return m_Call;}
	double GetPutPrice()	{return m_Put;}
	double GetCallDelta()	{return m_CallDelta;}
	double GetPutDelta()	{return m_PutDelta;}
	double GetGamma()		{return m_Gamma;}
	double GetVega()		{return m_Vega;}

protected:
// Inputs
	double m_Texp, m_Fwd, m_Strike, m_Vol;

// Outputs
	double m_Call, m_Put, m_Straddle, m_Intrinsic;
	double m_CallDelta, m_PutDelta;
	double m_Gamma, m_Vega;

// Interim calculations
	double m_RootT, m_vRootT;
};
