//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains miscellaneous examples relating to the use
// of the C API accessed via the Excel4() and Excel4v() functions.
//============================================================================
//============================================================================
#include <windows.h>
#include <stdio.h>
#include <math.h>

#include "XllAddIn.h"
#include "XllNames.h"
#include "TLS.h"

//============================================================================
char *Excel4_err_msg(int err_num)
{
	switch(err_num)
	{
	case xlretAbort:	return "Error: macro halted";
	case xlretInvXlfn:	return "Error: invalid function number";
	case xlretInvCount:	return "Error: invalid number of arguments";
	case xlretInvXloper: return "Error: invalid oper structure";
	case xlretStackOvfl: return "Error: stack overflow";
	case xlretUncalced:	return "Error: uncalced cell";
	case xlretFailed:	return "Error: command failed";
// Excel 12
	case xlretNotThreadSafe: return "Error: not thread safe";

	default: return "Unrecognised error";
	}
}
//============================================================================
xloper * __stdcall XLM4(int xlfn, xloper *arg0, xloper *arg1,
					   xloper *arg2, xloper *arg3, xloper *arg4,
					   xloper *arg5, xloper *arg6, xloper *arg7,
					   xloper *arg8, xloper *arg9, xloper *arg10,
					   xloper *arg11, xloper *arg12, xloper *arg13,
					   xloper *arg14, xloper *arg15, xloper *arg16,
					   xloper *arg17, xloper *arg18)
{
#pragma EXPORT

// Get a thread-local static xloper
	xloper *p_ret_val = get_thread_local_xloper();

	if(!p_ret_val) // Could not get a thread-local copy
		return NULL;

	const xloper *arg_array[19];

// Fill in array of pointers to the xloper arguments ready for the call
// to Excel4v()
	arg_array[0] = arg0;
	arg_array[1] = arg1;
	arg_array[2] = arg2;
	arg_array[3] = arg3;
	arg_array[4] = arg4;
	arg_array[5] = arg5;
	arg_array[6] = arg6;
	arg_array[7] = arg7;
	arg_array[8] = arg8;
	arg_array[9] = arg9;
	arg_array[10] = arg10;
	arg_array[11] = arg11;
	arg_array[12] = arg12;
	arg_array[13] = arg13;
	arg_array[14] = arg14;
	arg_array[15] = arg15;
	arg_array[16] = arg16;
	arg_array[17] = arg17;
	arg_array[18] = arg18;

// Find the last non-missing argument
	int i;
	for(i = 19; --i >= 0;)
		if((arg_array[i]->xltype & (xltypeMissing | xltypeNil)) == 0)
			break;

// Call the function. (i+1 because i counts from 0)
	int retval = Excel4v(xlfn, p_ret_val, i + 1, arg_array);

	if(retval != xlretSuccess)
	{
// If the call to Excel4v() failed, return a string explaining why
// and tell Excel to call back into the DLL to free the memory about
// to be allocated for the return string.
		p_ret_val->xltype = xltypeStr | xlbitDLLFree;
		p_ret_val->val.str = new_xlstring(Excel4_err_msg(retval));
	}
	else
	{
// Tell Excel Excel to free up memory that it might have allocated for
// the return value.
		if(p_ret_val->xltype & (xltypeStr | xltypeMulti | xltypeRef))
			p_ret_val->xltype |= xlbitXLFree;
	}
	return p_ret_val;
}
//============================================================================
xloper * __stdcall Excel4_match(xloper *p_lookup_value, xloper *p_lookup_array, int match_type)
{
#pragma EXPORT

// Get a thread-local static xloper
	xloper *p_ret_val = get_thread_local_xloper();

	if(!p_ret_val) // Could not get a thread-local copy
		return NULL;

// Convert the integer argument into an xloper so that a pointer
// to this can be passed to Excel4()
	xloper match_type_oper = {0.0, xltypeInt};
	match_type_oper.val.w = match_type;

	int xl4 = Excel4(
		xlfMatch,	// 1st arg: the function to be called
		p_ret_val,	// 2nd arg: ptr to return value
		3,			// 3rd arg: number of subsequent args
		p_lookup_value,	// fn arg1
		p_lookup_array,	// fn arg2
		&match_type_oper);// fn arg3

// Test the return value of Excel4()
	if(xl4 != xlretSuccess)
	{
		p_ret_val->xltype = xltypeErr;
		p_ret_val->val.err = xlerrValue;
	}
	else
	{
// Tell Excel to free up memory that it might have allocated for
// the return value.
		p_ret_val->xltype |= xlbitXLFree;
	}
	return p_ret_val;
}
#if 0 // simplified example
xloper * __stdcall Excel4_match(xloper *p_lkp_value,
			xloper *p_lkp_array, xloper *p_match_type)
{
#pragma EXPORT

	xloper *p_ret_val = get_thread_local_xloper();
	if(!p_ret_val) // Could not get a thread-local copy
		return NULL;

	Excel4(xlfMatch, p_ret_val, 3, p_lkp_value, p_lkp_array, p_match_type);
	p_ret_val->xltype |= xlbitXLFree;
	return p_ret_val;
}
#endif
//============================================================================
xloper * __stdcall cppExcel4_match(xloper *p_lookup_value, xloper *p_lookup_array, int match_type)
{
#pragma EXPORT

	cpp_xloper RetVal;
	xloper match_oper = {(double)match_type, xltypeNum};
// Excel is called here with xloper * arguments only
	RetVal.Excel(xlfMatch, 3, p_lookup_value, p_lookup_array, &match_oper);
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_cell(xloper *pRef, int arg_num)
{
#pragma EXPORT

	cpp_xloper RetVal, Ref(pRef), Arg(arg_num, 1, 66);
// Excel is called here with cpp_xloper * arguments only
	RetVal.Excel(xlfGetCell, 2, &Arg, &Ref); // Not thread-safe
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_document(int arg_num, char *sheet_name)
{
#pragma EXPORT

	cpp_xloper Arg1(arg_num, 1, 88);
	if(!Arg1.IsType(xltypeInt))
		return p_xlErrValue;

	cpp_xloper Arg2(sheet_name);
	cpp_xloper RetVal;
	RetVal.Excel(xlfGetDocument, 2, &Arg1, &Arg2); // Not thread-safe
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_def(char *DefText, char *DocText, int type)
{
#pragma EXPORT

	cpp_xloper Arg1(DefText);
	cpp_xloper Arg2(DocText);
	cpp_xloper Arg3(type, 1, 3);
	cpp_xloper RetVal;
	RetVal.Excel(xlfGetDef, 3, &Arg1, &Arg2, &Arg3);
	return RetVal.ExtractXloper();
}
//============================================================================
int __stdcall define_new_name(void)
{
#pragma EXPORT

// Get the name to be defined from the active cell. First get a
// reference to the active cell.  No need to evaluate it, as call
// to xlcDefineName will try to convert contents of cell to a
// string and use that.

	cpp_xloper ActiveCell, RetVal;
	if(ActiveCell.Excel(xlfActiveCell) == xlretSuccess)
		RetVal.Excel(xlcDefineName | xlPrompt, 1, &ActiveCell);

	return 1;
}
//============================================================================
xloper * __stdcall get_formula(xloper *p_ref)
{
#pragma EXPORT

	cpp_xloper RetVal;
	RetVal.Excel(xlfGetFormula, 1, p_ref);
// Extract and return the xloper, flagging Excel to free memory
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_name(char *name, xloper *p_info_type)
{
#pragma EXPORT

	cpp_xloper Name(name), InfoType(p_info_type);
	Name.Excel(xlfGetName, 2, &Name, &InfoType);
	return Name.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_note(long row, long column)
{
#pragma EXPORT

// Create a simple single-cell reference to a cell on the current sheet
	cpp_xloper RetVal, Arg((RW)row, (RW)row, (COL)column, (COL)column);
	RetVal.Excel(xlfGetNote, 1, &Arg);
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_window(int arg_num, char *window_name)
{
#pragma EXPORT

	cpp_xloper Arg1(arg_num, 1, 31);
	if(!Arg1.IsType(xltypeInt))
		return p_xlErrValue;

	cpp_xloper RetVal, Arg2(window_name);
	RetVal.Excel(xlfGetWindow, 2, &Arg1, &Arg2);
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_workbook(int arg_num, char *book_name)
{
#pragma EXPORT

	cpp_xloper Arg1(arg_num, 1, 38);
	if(!Arg1.IsType(xltypeInt))
		return p_xlErrValue;

	cpp_xloper RetVal, Arg2(book_name);
	RetVal.Excel(xlfGetWorkbook, 2, &Arg1, &Arg2);
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall name_array(char *WorkbookName, int NameType, char *Mask)
{
#pragma EXPORT

	cpp_xloper Arg1(WorkbookName);
	cpp_xloper Arg2(NameType, 1, 3);
	cpp_xloper Arg3(Mask);
	cpp_xloper RetVal;

	RetVal.Excel(xlfNames);
//	RetVal.Excel(xlfNames, 1, &Arg1);
//	RetVal.Excel(xlfNames, 2, &Arg1, &Arg2);
//	RetVal.Excel(xlfNames, 3, &Arg1, &Arg2, &Arg3);

// Extract and return xloper.
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall selection(int trigger)
{
#pragma EXPORT

	cpp_xloper RetVal;
	RetVal.Excel(xlfSelection);
// Extract and return xloper.
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall xl_windows(int match_type, char *mask)
{
#pragma EXPORT

	cpp_xloper Arg1(match_type, 1, 3);
	cpp_xloper Arg2(mask);
	cpp_xloper RetVal;
	RetVal.Excel(xlfWindows, 2, &Arg1, &Arg2);
// Extract and return xloper.
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall formula_convert(char *p_ref, int from_A1, int to_A1, int abs_rel_type, xloper *p_rel_ref)
{
#pragma EXPORT

	cpp_xloper Ref(p_ref), Arg4(abs_rel_type, 1, 4);
	cpp_xloper FromA1(from_A1 ? p_xlTrue : p_xlFalse);
	cpp_xloper ToA1(to_A1 ? p_xlTrue : p_xlFalse);
	cpp_xloper RelRef(p_rel_ref); // shallow copy if version < 12
	cpp_xloper RetVal;
// All arguments passed as cpp_xloper *
	RetVal.Excel(xlfFormulaConvert, 5, &Ref, &FromA1, &ToA1, &Arg4, &RelRef);
// Extract and return xloper.
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall text_ref(char *p_ref, int A1_style)
{
#pragma EXPORT

	cpp_xloper Op(p_ref);
	cpp_xloper A1Style(A1_style != 0); // Initialise to xltypeBool
// All arguments passed as cpp_xloper *
	Op.Excel(xlfTextref, 2, &Op, &A1Style);
// Extract and return xloper.
	return Op.ExtractXloper();
}
//============================================================================
xloper * __stdcall ref_text(xloper *p_ref, int A1_style)
{
#pragma EXPORT

	cpp_xloper RetVal;
// All arguments passed as xloper *
	xloper *pA1_style = A1_style != 0 ? p_xlTrue : p_xlFalse;
	RetVal.Excel(xlfReftext, 2, p_ref, pA1_style);
// Extract and return xloper.
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall get_workspace(int arg_num)
{
#pragma EXPORT

	cpp_xloper Op(arg_num, 1, 72);
	if(!Op.IsType(xltypeInt))
		return p_xlErrValue;
	Op.Excel(xlfGetWorkspace, 1, &Op); // Re-use Op
	return Op.ExtractXloper();

}
//============================================================================
xloper * __stdcall get_caller_info(int param)
{
#pragma EXPORT

	cpp_xloper Caller;
	cpp_xloper GetCell_param(param);
	cpp_xloper RetVal;

	Caller.Excel(xlfCaller);
	RetVal.Excel(xlfGetCell, 2, &GetCell_param, &Caller);

	return RetVal.ExtractXloper();
}
//============================================================================
#if 1 // Use the cpp_xloper class and use xlCoerce instead of xlfGetCell
xloper * __stdcall toggle_caller(void)
{
#pragma EXPORT
	cpp_xloper Caller;
	Caller.Excel(xlfCaller);
	if(!Caller.IsRef())
		return NULL;
	cpp_xloper TypeNum(xltypeNum);
	Caller.Excel(xlCoerce, 2, &Caller, &TypeNum);
	Caller = ((double)Caller == 0.0) ? 1.0 : 0.0;
	return Caller.ExtractXloper();
}
#else // use xlopers directly and use xlfGetCell instead of xlCoerce
xloper * __stdcall toggle_caller(void)
{
#pragma EXPORT

// Use of static here is not thread-safe, but function cannot be
// exported as thread-safe in any case since it must be registered
// as type # in order to be able to call xlfGetCell
	static xloper ret_val;
	xloper caller, GetCell_param;

	GetCell_param.xltype = xltypeInt;
	GetCell_param.val.w = 5; // contents of cell as number
	Excel4(xlfCaller, &caller, 0);
	Excel4(xlfGetCell, &ret_val, 2, &GetCell_param, &caller);
	if(ret_val.xltype == xltypeNum)
		ret_val.val.num = (ret_val.val.num == 0 ? 1.0 : 0.0);
	Excel4(xlFree, 0, 1, &caller);
	return &ret_val;
}
#endif
//============================================================================
xloper * __stdcall increment_caller(xloper *trigger)
{
#pragma EXPORT

	cpp_xloper Caller;
	Caller.Excel(xlfCaller); // Get a refence to the calling cell
	if(!Caller.IsRef())
		return NULL;
	Caller = (double)Caller + 1.0; // Coerce to number and increment
	return Caller.ExtractXloper();
}
//============================================================================
xloper * __stdcall app_title(xloper *title)
{
#pragma EXPORT

// Get a thread-local static xloper
	xloper *p_ret_val = get_thread_local_xloper();

	if(!p_ret_val) // Could not get a thread-local copy
		return NULL;

	Excel4(xlfAppTitle, p_ret_val, 1, title);
	return p_ret_val;
}
//============================================================================
char *get_dll_name1(void)
{
	cpp_xloper DllName;
// Return a deep copy of the string (needs to be freed by the caller)
	return DllName.Excel(xlGetName) == xlretSuccess && DllName.IsStr()
 		? (char *)DllName : NULL;
}
//============================================================================
int __stdcall show_dll_name(void)
{
#pragma EXPORT

	cpp_xloper DllName;
	if(DllName.Excel(xlGetName) != xlretSuccess)
		return 0;

	DllName.Alert();
	return 1;
}
//============================================================================
char *get_dll_name2(void)
{
	xloper dll_name;
	if(Excel4(xlGetName, &dll_name, 0) != xlretSuccess
	|| dll_name.xltype != xltypeStr)
		return NULL;

// Make a copy of the string (needs to be freed by the caller)
	size_t len = (BYTE)dll_name.val.str[0];
	char *name = (char *)malloc(len + 1);

	memcpy(name, dll_name.val.str + 1, len);
	name[len] = 0;
	Excel4(xlFree, 0, 1, &dll_name);
	return name;
}
//============================================================================
xloper * __stdcall sheet_id(char *sheetname)
{
#pragma EXPORT

	cpp_xloper SheetId;
	cpp_xloper Name(sheetname);

	SheetId.Excel(xlSheetId, 1, &Name);

	if(!SheetId.IsType(xltypeRef))
		return p_xlErrValue;

	double id;
	if(gExcelVersion12plus)
		id = SheetId.OpAddr12()->val.mref.idSheet;
	else
		id = SheetId.OpAddr()->val.mref.idSheet;

	cpp_xloper RetVal(id);
	return RetVal.ExtractXloper(); // Converts xloper12 down to xloper if required
}
//============================================================================
xloper * __stdcall sheet_name(double ID)
{
#pragma EXPORT

// Get a thread-local static xloper
	xloper *p_ret_val = get_thread_local_xloper();

	if(!p_ret_val) // Could not get a thread-local copy
		return NULL;

	xloper ID_ref_oper;

	if(ID < 0)
	{
		ID_ref_oper.xltype = xltypeNil;
	}
	else
	{
		ID_ref_oper.xltype = xltypeRef;
		ID_ref_oper.val.mref.idSheet = (DWORD)ID;
		ID_ref_oper.val.mref.lpmref = NULL;
	}
	Excel4(xlSheetNm, p_ret_val, 1, &ID_ref_oper);
	p_ret_val->xltype |= xlbitXLFree;
	return p_ret_val;
}
//============================================================================
xloper * __stdcall sheet_name_cpp(double ID)
{
#pragma EXPORT

	xloper ID_ref_oper;

	if(ID < 0)
	{
		ID_ref_oper.xltype = xltypeNil;
	}
	else
	{
		ID_ref_oper.xltype = xltypeRef;
		ID_ref_oper.val.mref.idSheet = (DWORD)ID;
		ID_ref_oper.val.mref.lpmref = NULL;
	}

	cpp_xloper SheetNm;
	SheetNm.Excel(xlSheetNm, 1, &ID_ref_oper);
	return SheetNm.ExtractXloper();
}
//============================================================================
bool register_example(void)
{
	cpp_xloper DllName;
	cpp_xloper FunctionName("exponent_function");
	cpp_xloper TypeText("BB"); // = return a double, take a double
	cpp_xloper WorksheetFunctionName("MY_EXP");
	cpp_xloper Arguments("Exponent");
	cpp_xloper FunctionType(1);
	cpp_xloper Category("My functions");
	cpp_xloper Description("Returns e to the power of Exponent");
	cpp_xloper Arg1Help("Any number such that |n| <= 709");
	cpp_xloper Missing(p_xlMissing);
	cpp_xloper RetVal;

// Get the full path and name of the DLL.
	if(DllName.Excel(xlGetName) != xlretSuccess)
		return false;

	int ret_val = RetVal.Excel(xlfRegister,
		11, // number of subsequent arguments
		&DllName,
		&FunctionName,
		&TypeText,
		&WorksheetFunctionName,
		&Arguments,
		&FunctionType,
		&Category,
		&Missing, // no short-cut
		&Missing, // no help topic
		&Description,
		&Arg1Help);

	if(ret_val != xlretSuccess)
	{
		cpp_xloper Message("Could not register MY_EXP");
		Message.Alert();
		return false;
	}
	return true;
}
//============================================================================
double __stdcall exponent_function(double arg1)
{
#pragma EXPORT

	if(fabs(arg1) > 709) // limit of e^arg1 in IEEE double
		return 0.0;

	return exp(arg1);
}
//============================================================================
#if 1
double __stdcall function_break_example(xloper *arg)
{
#pragma EXPORT

	if(arg->xltype != xltypeNum)
		return -1;

	cpp_xloper BreakState;
	long l;
	for(l = (long)arg->val.num; --l;)
	{
// Detect a user break attempt but leave it set so that other
// worksheet functions can also detect it
		BreakState.Excel(xlAbort);

		if(BreakState.IsTrue())
			break;
	}
	return l;
}
//============================================================================
int __stdcall command_break_example(void)
{
	cpp_xloper BreakState, False(false);
	for(;;)
	{
// Detect a user break and then clear it before exiting
		BreakState.Excel(xlAbort);

		if(BreakState.IsTrue()) // then reset it
		{
			BreakState.Excel(xlAbort, 1, &False);
			return 0;
		}
	}
	return 1;
}
#else
//============================================================================
double __stdcall function_break_example(xloper *arg)
{
#pragma EXPORT

	if(arg->xltype != xltypeNum)
		return -1;

	long l;

	if(gExcelVersion12plus)
	{
		xloper12 break_state;

		for(l = (long)arg->val.num; --l;)
		{
// Detect a user break attempt but leave it set so that other
// worksheet functions can also detect it
			Excel12(xlAbort, &break_state, 0);

			if(break_state.xltype == xltypeBool && break_state.val.xbool == 1)
				break;
		}
	}
	else
	{
		xloper break_state;

		for(l = (long)arg->val.num; --l;)
		{
// Detect a user break attempt but leave it set so that other
// worksheet functions can also detect it
			Excel4(xlAbort, &break_state, 0);

			if(break_state.xltype == xltypeBool && break_state.val.xbool == 1)
				break;
		}
	}
	return l;
}
#endif
//============================================================================
double __stdcall get_stack(void)
{
#pragma EXPORT
	if(gExcelVersion12plus)
	{
		xloper12 retval;
		if(xlretSuccess != Excel12(xlStack, &retval, 0))
			return -1.0;

		if(retval.xltype == xltypeInt)
			return (double)retval.val.w; // returns min(64Kb, actual stack space)
// Microsoft state that this is not the returned type, but was returned in
// an Excel 12 beta, so the code is left here
		if(retval.xltype == xltypeNum)
			return retval.val.num;
	}
	else
	{
		xloper retval;
		if(xlretSuccess != Excel4(xlStack, &retval, 0))
			return -1.0;

		if(retval.xltype == xltypeInt)
			return (double)(unsigned short)retval.val.w;
	}
	return -1.0;
}
//============================================================================
int __stdcall xl_call_version(void)
{
#pragma EXPORT
	cpp_xloper Version(XLCallVer()); // returns an integer
	Version = ((int)Version) / 265;
	Version.Alert(); // convert integer to string and display in dialog
	return 1;
}
//============================================================================
xloper * __stdcall evaluate(xloper *p_formula) // exported as worksheet equivalent
{
#pragma EXPORT

	cpp_xloper RetVal;
	RetVal.Excel(xlfEvaluate, 1, p_formula);
	return RetVal.ExtractXloper();
}
//============================================================================
xloper * __stdcall evaluate2(xloper *p_formula) // exported as macro-sheet equivalent
{
#pragma EXPORT

	cpp_xloper RetVal;
	RetVal.Excel(xlfEvaluate, 1, p_formula);
	return RetVal.ExtractXloper();
}
//============================================================================
void display_evaluate_output(cpp_xloper &Formula)
{
	cpp_xloper RetVal, Type, MsgString;
	RetVal.Excel(xlfEvaluate, 1, &Formula);
	Type = RetVal.GetType();

	MsgString = "Formula: ";
	MsgString += Formula;
	MsgString += ", Value: ";
	MsgString += RetVal;
	MsgString += ", Type: ";
	MsgString += Type;
	MsgString.Alert();
}
//----------------------------------------------------------------------------
int __stdcall evaluate3(void) // exported as command
{
#pragma EXPORT
	cpp_xloper RunThis("Evaluates some test range formulae and displays the result and type.");
	if(RunThis.Alert(1) == false) // User pressed Cancel button
		return 0;

	cpp_xloper Formula("TestCell");
	display_evaluate_output(Formula);

	Formula = "!TestCell";
	display_evaluate_output(Formula);

	Formula = "Sheet1!TestCell";
	display_evaluate_output(Formula);

	Formula = "DllTestName";
	display_evaluate_output(Formula);
	return 1;
}
//============================================================================
bool excel_using_1904_system(void)
{
	cpp_xloper Using1904; // initialised to xltypeNil
	cpp_xloper Arg(20); // initialised to xltypeInt
	Using1904.Excel(xlfGetDocument, 1, &Arg);
	return Using1904.IsTrue();
}
// This version is quicker as the creator/destructor
// calls and the overloaded operator calls are not
// required.  An optimising compiler might produce code
// this efficient, of course.
bool excel_using_1904_system_c(void)
{
	xloper using_1904;
	xloper arg;
	arg.xltype = xltypeInt;
	arg.val.w = 20;
	Excel4(xlfGetDocument, &using_1904, 1, &arg);
	return using_1904.xltype == xltypeBool && using_1904.val.xbool == 1;
}
//============================================================================
#define DAYS_1900_TO_1904	1462 // = 1-Jan-1904 in 1900 system

int __stdcall worksheet_date_fn(int input_date)
{
#pragma EXPORT

	bool using_1904 = excel_using_1904_system();

	if(using_1904)
		input_date += DAYS_1900_TO_1904;

// Do something with the date
	int result = 0; // some_date_fn(input_date);

	if(using_1904)
		result -= DAYS_1900_TO_1904;

	return result;
}
//============================================================================
xloper * __stdcall count_used_cells(int first_row, int last_row, int first_col, int last_col)
{
#pragma EXPORT

	if(first_row > last_row || first_col > last_col)
		return p_xlErrValue;

// Adjust inputs to be zero-counted and cast to RWs and COLs.  (Casts
// are not strictly necessary as RW and COL are defined as 32-bit ints)
	RW fr = (RW)(first_row - 1);
	RW lr = (RW)(last_row - 1);
	COL fc = (COL)(first_col - 1);
	COL lc = (COL)(last_col - 1);
	cpp_xloper Op(fr, lr, fc, lc);
	Op.Excel(xlfCount, 1, &Op); // re-use Op
	return Op.ExtractXloper();
}
//============================================================================
xloper * __stdcall count_used_cells2(char *sheetname, int first_row, int last_row, int first_col, int last_col)
{
#pragma EXPORT

	if(first_row > last_row || first_col > last_col)
		return p_xlErrValue;

// Adjust inputs to be zero-counted and cast to RWs and COLs.  (Casts
// are not strictly necessary as RW and COL are defined as 32-bit ints)
	RW fr = (RW)(first_row - 1);
	RW lr = (RW)(last_row - 1);
	COL fc = (COL)(first_col - 1);
	COL lc = (COL)(last_col - 1);
	cpp_xloper Op(sheetname, fr, lr, fc, lc);
	Op.Excel(xlfCount, 1, &Op); // re-use Op
	return Op.ExtractXloper();
}
//============================================================================
double __stdcall  C_indirect1(long trigger, long row, long col)
{
#pragma EXPORT

// Convert to zero-base and right types for constructor
	RW r = (RW)(row - 1);
	COL c = (COL)(col - 1);

// Create xltypeSref from inputs and coerce to double
	cpp_xloper Source(r, r, c, c);
	return (double)Source;
}
//============================================================================
double __stdcall  C_indirect2(long trigger, xloper *p_source)
{
#pragma EXPORT

	cpp_xloper Source(p_source);
	cpp_xloper ResultType(xltypeNum); // Get Excel to try and convert to a f.p. number
	cpp_xloper Result;

	if(Result.Excel(xlCoerce, 2, &Source, &ResultType) != xlretSuccess)
		return 0.0;

	return (double)Result;
}
//============================================================================
void show_alert_with_arg(char *msg, char *arg)
{
	if(!msg || !arg)
		return;

	cpp_xloper Message(msg);
	Message += arg; // string concatenation
	Message.Alert();
}
//============================================================================
bool get_calling_cell_value(xloper &value)
{
//-----------------------------------------------------------
// Get a reference to the calling cell
//-----------------------------------------------------------
	xloper caller;
	int xl4 = Excel4(xlfCaller, &caller, 0);

	if(xl4) // if xlfCaller failed
		return false;

	if(!(caller.xltype & (xltypeRef | xltypeSRef)))
	{
		Excel4(xlFree, 0, 1, &caller);
		return false;
	}

//-----------------------------------------------------------
// Get the calling cell's value
//-----------------------------------------------------------

	xloper xl_ret_val;

	xl4 = Excel4(xlCoerce, &xl_ret_val, 1, &caller);

	if(xl4)
	{
		Excel4(xlFree, 0, 1, &caller);
		return false;
	}

//-----------------------------------------------------------
// Clone the value and free the Excel-allocated memory
//-----------------------------------------------------------
	if(!clone_xloper(&value, &xl_ret_val))
	{
		Excel4(xlFree, 0, 2, &caller, &xl_ret_val);
		return false;
	}

	Excel4(xlFree, 0, 2, &caller, &xl_ret_val);
	return true;
}
//============================================================================
#define USE_CPP_XLOPER 1

#if !USE_CPP_XLOPER

// This function is used as an example of a function where
// allocated memory needs to be freed by the DLL after
// having been returned to Excel.  (See section 7.4)

xloper * __stdcall random_array(int rows, int columns)
{
#pragma EXPORT

// Get a thread-local static xloper
	xloper *p_ret_val = get_thread_local_xloper();
	
	if(!p_ret_val) // Could not get a thread-local copy
		return NULL;

	int array_size = rows * columns;
	xloper *array;

	if(array_size <= 0)
		return NULL;

	array = (xloper *)malloc(array_size * sizeof(xloper));
	if(array == NULL)
		return NULL;

	for(int i = array_size; --i >= 0;)
		Excel4(xlfRand, array + i, 0); 

// Instruct Excel to call back into DLL to free the memory
	p_ret_val->xltype = xltypeMulti | xlbitDLLFree;
	p_ret_val->val.array.lparray = array;
	p_ret_val->val.array.rows = rows;
	p_ret_val->val.array.columns = columns;
	return p_ret_val;
}

#else

xloper * __stdcall random_array(int rows, int columns)
{
#pragma EXPORT

	cpp_xloper array((RW)rows, (COL)columns);

	if(!array.IsType(xltypeMulti))
		return NULL;

	DWORD array_size;
	array.GetArraySize(array_size);
	cpp_xloper ArrayElt;

	for(DWORD i = 0; i < array_size; i++)
	{
		if(array.GetArrayElt(i, ArrayElt))
		{
			ArrayElt.Excel(xlfRand);
			array.SetArrayElt(i, ArrayElt);
		}
	}
	return array.ExtractXloper();
}

#endif
//============================================================================

HRESULT AutoWrap(int autoType, VARIANT *pvResult, IDispatch *pDisp, LPOLESTR ptName, int cArgs...)
{
// Begin variable-argument list...
    va_list marker;
    va_start(marker, cArgs);

    if(!pDisp)
	{
        MessageBox(NULL, "NULL IDispatch passed to AutoWrap()", "Error", 0x10010);
        _exit(0);
    }

// Variables used...
    DISPPARAMS dp = { NULL, NULL, 0, 0 };
    DISPID dispidNamed = DISPID_PROPERTYPUT;
    DISPID dispID;
    HRESULT hr;
    char buf[200];
    char szName[200];
    
// Convert down to ANSI
    WideCharToMultiByte(CP_ACP, 0, ptName, -1, szName, 256, NULL, NULL);
    
// Get DISPID for name passed...
    hr = pDisp->GetIDsOfNames(IID_NULL, &ptName, 1, LOCALE_USER_DEFAULT, &dispID);
    if(FAILED(hr))
	{
        sprintf(buf, "IDispatch::GetIDsOfNames(\"%s\") failed w/err 0x%08lx", szName, hr);
        MessageBox(NULL, buf, "AutoWrap()", 0x10010);
        _exit(0);
        return hr;
    }
    
// Allocate memory for arguments...
    VARIANT *pArgs = new VARIANT[cArgs+1];
// Extract arguments...
    for(int i=0; i<cArgs; i++)
	{
        pArgs[i] = va_arg(marker, VARIANT);
    }

// Build DISPPARAMS
    dp.cArgs = cArgs;
    dp.rgvarg = pArgs;

// Handle special-case for property-puts!
    if(autoType & DISPATCH_PROPERTYPUT)
	{
        dp.cNamedArgs = 1;
        dp.rgdispidNamedArgs = &dispidNamed;
    }

// Make the call!
    hr = pDisp->Invoke(dispID, IID_NULL, LOCALE_SYSTEM_DEFAULT, autoType, &dp, pvResult, NULL, NULL);
    if(FAILED(hr))
	{
        sprintf(buf, "IDispatch::Invoke(\"%s\"=%08lx) failed w/err 0x%08lx", szName, dispID, hr);
        MessageBox(NULL, buf, "AutoWrap()", 0x10010);
        _exit(0);
        return hr;
    }
// End variable-argument section...
    va_end(marker);
    delete [] pArgs;
    return hr;
}
//============================================================================
#if 0
int test_range_assignment(void)
{
    VARIANT root[64] = {0}; // Generic IDispatchs
    VARIANT parm[64] = {0}; // Generic Parameters
    VARIANT rVal = {0}; // Temporary result holder
    int level=0; // Current index into root[]

    // Initialize the OLE Library...
    OleInitialize(NULL);

    // Line 1: Dim r As Range, s As String
    VARIANT r = {0}, s = {0};

    // Line 4: Set r = Range A1 
    VariantCopy(&rVal, &Range);
    VariantCopy(&r, &rVal);
    VariantClear(&rVal);

    // Line 5: s = r . Value 
    VariantCopy(&root[++level], &r);
    AutoWrap(DISPATCH_PROPERTYGET|DISPATCH_METHOD, &rVal, root[level].pdispVal, L"Value", 0);
    VariantClear(&root[level--]);
    VariantCopy(&s, &rVal);
    VariantClear(&rVal);

    // Clearing variables
    VariantClear(&r);
    VariantClear(&s);

    // Close the OLE Library...
    OleUninitialize();
}
#endif