//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file (and header file) contain the definitions of various
// static structures and strings.  The structures contain the data passed to
// Excel when registering commands and functions.
//============================================================================
//============================================================================
#include <windows.h>

#include "XllAddIn.h"
#include "XllNames.h"

char *AddinVersionStr = "EADC (2nd edn) example add-in V1.0";
char *AddinName = "EADC Fns";


//---------------------------------------------------------
// Commands
//---------------------------------------------------------
char *CommandExports[NUM_COMMANDS][2] =
{
//	Name in code			Name that Excel uses
	{"define_new_name",		"DefineNewName"},
	{"clean_xll_name_list",	"CleanNameList"},
	{"xl_call_version",		"XLCallVersion"},
	{"on_time_example_cmd",	"OnTimeExample"},
	{"increment_counter",	"IncrementCounter"},
	{"show_example_dialog",	"ShowExampleDialog"},
	{"get_username",		"GetUsername"},
	{"LongTaskPollingCmd",	"LongTaskPoll"},
	{"LongTaskToggleCmd",	"LongTaskToggleCmd"},
	{"LongTaskConfigCmd",	"LongTaskConfig"},
	{"monte_carlo_control",	"MC_Control"},
	{"set_values",			"SetValues"},
	{"evaluate3",			"EvaluateCmd"},
};

//---------------------------------------------------------
//---------------------------------------------------------
// Exported worksheet function information
//---------------------------------------------------------
//---------------------------------------------------------

//---------------------------------------------------------
// DUAL-INTERFACE METHOD:
//
// struct ws_func_export_data supports both an Excel 11-
// and an Excel 12+ interface to be defined.  If running
// under Excel 11- the Excel 11- interface is registered.
// If running under Excel 12+ the Excel 12+ interface is
// registered if it has been defined, otherwise the Excel
// 11- interface is registered instead.
//---------------------------------------------------------

// Arguments for register_ws_function() which calls xlfRegister
// to register worksheet functions with Excel.

// Arg'n' denotes the number of the argument passed to xlfRegister.
// Arg1 is the DLL path and file name. It is the same for all exports
// and is populated in register_ws_function().

// If name_in_code12 is omitted, arg_names12 is ignored and
// (even when running with Excel 12+) the Excel 11- interface is
// registered. In this case, if types12 is given, it will be used
// as this permits the same function to registered as thread-safe
// in Excel12.

// If name_in_code12 is provided and running with Excel 12+ types12
// is required.  If arg_names12 is not supplied, arg_names is used,
// and the Excel 11- interface is not registered.

// 11 args preceed 1st arg_help in calls to xlfRegister, so at most can only
// supply MAX_XL12_UDF_ARGS - 11 help text strings in Excel 12.  If running
// with Excel 11- only the first MAX_XL11_UDF_ARGS - 11 of these will be used

#if 0 // Already defined. Listed here for reference only.
typedef struct
{
// REQUIRED (see comments above):
	char *name_in_code;		// Arg2: Function name as in code (v11-)
	char *types;			// Arg3: Return type and argument types (v11-)
	char *name_in_code12;	// Arg2: Function name as in code (v12+)
	char *types12;			// Arg3: Return type and argument types (v12+)
	char *ws_name;			// Arg4: Fn name as it appears on worksheet
	char *arg_names;		// Arg5: Argument names (Excel 11-: max 30)
	char *arg_names12;		// Arg5: Argument names (Excel 12+: max 255)
// OPTIONAL:
	char *fn_category;		// Arg7: Function category for Function Wizard
	char *help_file;		// Arg9: Help file (optional)
	char *fn_description;	// Arg10: Function description text (optional)
	char *arg_help[MAX_XL12_UDF_ARGS - 11]; // Arg11...: Fn arg help text strings
}
	dual_ws_func_export_data;
#endif

dual_ws_func_export_data DualWsFuncExports[NUM_DUAL_IFACE_FUNCTIONS] =
{
//---------------------------------------------------------
// Chapter 10: String functions.
//---------------------------------------------------------
	{
		"concat_xl4",
		"RPPPPP",
		"concat_xl12",
		"UQQQQQ$",
		"Concat",
		"Input,Delim,MaxLen,NumDecs,NumScaling",
		"", // arg_names12
		"Text",
		"", // help file
		"Converts and concatentates cell contents to a delimted string",
		"Input range",
		"(optional) delimiter. Defaults to comma",
		"(optional) max output length. Defaults to 255 (v11-) or 32767 (v12+)",
		"(optional) num decimal places. Defaults to no rounding",
		"(optional) scaling multiplier for numbers  "
	},
	{
		"parse_xl4",
		"RCPPPP",
		"parse_xl12",
		"UC%QQQQ$",
		"ParseText",
		"Input,Delim,Numeric,EmptyVal,NumScaling",
		"", // arg_names12
		"Text",
		"", // help file
		"Converts delimted string to range of values",
		"Input range",
		"(optional) delimiter. Defaults to comma",
		"(optional) if TRUE, converts tokens to numbers",
		"(optional) the value to be placed in empty fields",
		"(optional) scaling multiplier for numbers  "
	},
	{
		"count_char_xl4",
		"HCP",
		"count_char_xl12",
		"HC%Q$",
		"CountChar",
		"Text,Char",
		"", // arg_names12
		"Text",
		"", // help file
		"Count number of occurances of char in text",
		"Text",
		"Char ",
	},
	{
		"replace_mask_xl4",
		"1FCP",
		"replace_mask_xl12",
		"1F%C%Q$",
		"ReplaceMask",
		"Text,OldMask,NewMask",
		"", // arg_names12
		"Text",
		"", // help file
		"Replace all occurances of OldMask chars with corresponding NewMask chars in text",
		"Text",
		"OldMask",
		"NewMask ",
	},
	{
		"reverse_text_xl4",
		"1F",
		"reverse_text_xl12",
		"1F%$",
		"ReverseText",
		"Text",
		"", // arg_names12
		"Text",
		"", // help file
		"Reverse text",
		"Text ",
	},
	{
		"find_first_xl4",
		"HCC",
		"find_first_xl12",
		"HC%C%$",
		"FindFirst",
		"Text,Mask",
		"", // arg_names12
		"Text",
		"", // help file
		"Find position of first char in text that's also in mask, 0 if not found",
		"Text",
		"Mask ",
	},
	{
		"find_first_excluded_xl4",
		"HCC",
		"find_first_excluded_xl12",
		"HC%C%$",
		"FindFirstExcl",
		"Text,Mask",
		"", // arg_names12
		"Text",
		"", // help file
		"Find position of first char in text that's NOT in mask, 0 if not found",
		"Text",
		"Mask ",
	},
	{
		"find_last_xl4",
		"HCP",
		"find_last_xl12",
		"HC%Q$",
		"FindLast",
		"Text,Mask",
		"", // arg_names12
		"Text",
		"", // help file
		"Find position of last char in text that's also in mask, 0 if not found",
		"Text",
		"Mask",
	},
	{
		"compare_text_xl4",
		"RCCP",
		"compare_text_xl12",
		"UC%C%Q$",
		"CompareText",
		"Atext,Btext,CaseSense",
		"", // arg_names12
		"Text",
		"", // help file
		"Compare strings, 0:same, <0:A<B, >0:A>B",
		"first string to compare",
		"second string to compare",
		"(Optional) False/omitted: case-insensitive  ",
	},
	{	
		"compare_nchars_xl4",
		"RCCHP",
		"compare_nchars_xl12",
		"UC%C%HQ$",
		"CompareNchars",
		"Atext,Btext,Nchars,CaseSense",
		"", // arg_names12
		"Text",
		"", // help file
		"Compare 1st n chars of strings, 0=same, -1=A<B , 1=A>B",
		"first string to compare",
		"second string to compare",
		"Nchars",
		"(Optional) True/omitted: case-sensitive  ",
	},
//---------------------------------------------------------
// Chapter 10: Probability functions
//-----------------------------------------------------------
	{
		"ndist",
		"BB",
		"ndist",
		"BB$",
		"Ndist",
		"x",
		"", // arg_names12
		"Statistical",
		"", // help file
		"Cumulative normal distribution (uses Excel NORMSDIST if version > 9)",
		"the value of the random variable  ",
	},
	{
		"ndist_inv_xl4",
		"RB",
		"ndist_inv_xl12",
		"UB$",
		"NdistInv",
		"Probability",
		"", // arg_names12
		"Statistical",
		"", // help file
		"Calculate the inverse of Ndist (uses Excel NORMSINV if version > 9)",
		"probability p, 0 > p > 1  ",
	},
	{
		"nsample_BM",
		"B!",
		"nsample_BM",
		"B!$",
		"NsampleBoxMuller",
		"", // arg_names
		"", // arg_names12
		"Statistical",
		"", // help file
		"Returns a normal distribution sample.  ",
	},
	{
		"nsample_BM_pair",
		"1K",
		"nsample_BM_pair",
		"1K$",
		"NsampleBoxMullerPair",
		"RandInput",
		"", // arg_names12
		"Statistical",
		"", // help file
		"Returns 2 normal distribution samples in the same shape as the input.",
		"1x2 or 2x1 array of random numbers in the range (0,1]  ",
	},
	{
		"xll_rand_xl4",
		"BPPP!",
		"xll_rand_xl12",
		"BQQQ!$",
		"RandXll",
		"Seed1,Seed2,Seed3",
		"", // arg_names12
		"Statistical",
		"", // help file
		"Returns a random number using same algorithm as Excel 2003 (Wichman-Hill).",
		"(Optional) a number between 1 and 30000",
		"(Optional) a number between 1 and 30000",
		"(Optional) a number between 1 and 30000",
	},
	{
		"xll_rand_non_vol_xl4",
		"BPPPP!",
		"xll_rand_non_vol_xl12",
		"BQQQQ!$",
		"RandXllnv",
		"Trigger,Seed1,Seed2,Seed3",
		"", // arg_names12
		"Statistical",
		"", // help file
		"Non volatile: Returns a random number using same algorithm as Excel 2003 (Wichman-Hill).",
		"(Optional) a number between 1 and 30000",
		"(Optional) a number between 1 and 30000",
		"(Optional) a number between 1 and 30000",
	},
//---------------------------------------------------------
// Chapter 10: Matrix functions
//-----------------------------------------------------------
	{
		"eigen_system_xl4",
		"RK",
		"eigen_system_xl4", // calls eigen_system_xl4 in v12
		"RK$",
		"EigenSystem",
		"Matrix",
		"", // arg_names12
		"Example",
		"",
		"Given NxN input, returns (N+1)xN evects and evals",
		"a square symetrical matrix ",
	},
//---------------------------------------------------------
// Chapter 10: Interpolation functions
//---------------------------------------------------------
	{
		"interp_xl4",
		"RBKKP",
		"interp_xl12",
		"UBKKQ$",
		"Interp",
		"X,Y_array,X_array,DontExtrapolate",
		"", // arg_names12
		"Example",
		"",
		"Interpolates F(X) from ranges of X and F values",
		"X lookup value",
		"row or column of Y values",
		"row or column of X values corresponding to Y values",
		"inhibits extrapolation if true or non-zero ",
	},
	{
		"make_spline_xl4",
		"RKBBJ",
		"make_spline_xl12",
		"UKBBJ$",
		"MakeSpline",
		"Input,StartGrad,EndGrad,Mode",
		"", // arg_names12
		"Example",
		"",
		"Creates a splined array of 2nd derivatives",
		"Input",
		"StartGrad",
		"EndGrad",
		"Gradients used: 0(dflt): neither, 1: start only, 2: end only, 3: both  ",
	},
	{
		"spline_interp_xl4",
		"RBKP",
		"spline_interp_xl12",
		"RBKP",
		"SplineInterp",
		"Input,X,Blend",
		"", // arg_names12
		"Example",
		"",
		"Interpolates along a spline",
		"Input",
		"X",
		"blend of 100% linear (0) to 100% cubic spline (1) ",
	},
//---------------------------------------------------------
// Chapter 10: 
//---------------------------------------------------------
	{
		"match_multi_xl4",
		"RPPPPPPPPPP",
		"match_multi_xl12",
		"UQQQQQQQQQQ$",
		"MatchMulti",
		"Value1,Range1,Value2,Range2,Value3,Range3,Value4,Range4,Value5,Range5",
		"", // arg_names12
		"Example",
		"",
		"Returns an offset where all the values match the entries in their corresponding ranges",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"",
	},
	{
		"sum_if_multi_xl4",
		"RPPPPPPPPPPP",
		"sum_if_multi_xl12",
		"UQQQQQQQQQQQ$",
		"SumIfMulti",
		"SumRange,Value1,Range1,Value2,Range2,Value3,Range3,Value4,Range4,Value5,Range5",
		"", // arg_names12
		"Example",
		"",
		"Sums the values of cells in sum range where all search values match the entries in the corresponding ranges",
		"the sum whose values are summed if search values match",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"",
	},
	{
		"count_if_multi_xl4",
		"RPPPPPPPPPP",
		"count_if_multi_xl12",
		"UQQQQQQQQQQ$",
		"CountIfMulti",
		"Value1,Range1,Value2,Range2,Value3,Range3,Value4,Range4,Value5,Range5",
		"", // arg_names12
		"Example",
		"",
		"Counts the number of cases where all search values match the entries in the corresponding ranges",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"",
	},
	{
		"average_if_multi_xl4",
		"RPPPPPPPPPPP",
		"average_if_multi_xl12",
		"UQQQQQQQQQQQ$",
		"AverageIfMulti",
		"AverageRange,Value1,Range1,Value2,Range2,Value3,Range3,Value4,Range4,Value5,Range5",
		"", // arg_names12
		"Example",
		"",
		"Averages the values of cells in sum range where all search values match the entries in the corresponding ranges",
		"the sum whose values are summed if search values match",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"the value to be searched for",
		"the range to search in",
		"",
	},
};

//============================================================================
//============================================================================
// Core function
bool concat_xl(cpp_xloper &RetVal, const cpp_xloper &Inputs, const cpp_xloper &Delim,
	const cpp_xloper &MaxLen, const cpp_xloper &NumDecs, const cpp_xloper &NumScaling);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers
xloper * __stdcall concat_xl4(xloper *inputs, xloper *p_delim,
	xloper *p_max_len, xloper *p_num_decs, xloper *p_num_scaling)
{
#pragma EXPORT

	cpp_xloper RetVal, Inputs(inputs), Delim(p_delim), MaxLen(p_max_len),
		NumDecs(p_num_decs), NumScaling(p_num_scaling);
	concat_xl(RetVal, Inputs, Delim, MaxLen, NumDecs, NumScaling);
	return RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s
xloper12 * __stdcall concat_xl12(xloper12 *inputs, xloper12 *p_delim,
	xloper12 *p_max_len, xloper12 *p_num_decs, xloper12 *p_num_scaling)
{
#pragma EXPORT

	cpp_xloper RetVal, Inputs(inputs), Delim(p_delim), MaxLen(p_max_len),
		NumDecs(p_num_decs), NumScaling(p_num_scaling);
	concat_xl(RetVal, Inputs, Delim, MaxLen, NumDecs, NumScaling);
	return RetVal.ExtractXloper12();
}
//============================================================================
// Core function
bool parse_xl(cpp_xloper &RetVal, const cpp_xloper &Input, const cpp_xloper &Delim,
	const cpp_xloper &Numeric, const cpp_xloper &Empty, const cpp_xloper &NumScaling);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
xloper * __stdcall parse_xl4(char *input, xloper *p_delim, xloper *p_numeric,
	xloper *p_empty, xloper *p_num_scaling)
{
#pragma EXPORT

	cpp_xloper RetVal, Input(input), Delim(p_delim), Numeric(p_numeric),
		Empty(p_empty), NumScaling(p_num_scaling);
	parse_xl(RetVal, Input, Delim, Numeric, Empty, NumScaling);
	return RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
xloper12 * __stdcall parse_xl12(wchar_t *input, xloper12 *p_delim, xloper12 *p_numeric,
	xloper12 *p_empty, xloper12 *p_num_scaling)
{
#pragma EXPORT

	cpp_xloper RetVal, Input(input), Delim(p_delim), Numeric(p_numeric),
		Empty(p_empty), NumScaling(p_num_scaling);
	parse_xl(RetVal, Input, Delim, Numeric, Empty, NumScaling);
	return RetVal.ExtractXloper12();
}
//============================================================================
// Core functions
size_t count_char(char *text, char ch);
size_t count_char(wchar_t *text, wchar_t ch);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
size_t __stdcall count_char_xl4(char *text, xloper *p_ch)
{
#pragma EXPORT

	cpp_xloper Ch(p_ch);
	char ch;
	if(Ch.IsStr())
		ch = (char)Ch.First();
	else if(Ch.IsNum())
		ch = (char)(double)Ch;
	else
		return 0;

	return count_char(text, ch);
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
size_t __stdcall count_char_xl12(wchar_t *text, xloper12 *p_ch)
{
#pragma EXPORT

	cpp_xloper Ch(p_ch);
	wchar_t ch;
	if(Ch.IsStr())
		ch = Ch.First();
	else if(Ch.IsNum())
		ch = (wchar_t)(double)Ch;
	else
		return 0;

	return count_char(text, ch);
}
//============================================================================
// Core functions
void replace_mask(char *text, char *old_chars, char *new_chars);
void replace_mask(wchar_t *text, wchar_t *old_chars, wchar_t *new_chars);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
void __stdcall replace_mask_xl4(char *text, char *old_chars, xloper *p_new_chars)
{
#pragma EXPORT

	cpp_xloper NewChars(p_new_chars);
	char *new_chars = NewChars.IsStr() ? (char *)NewChars : NULL;
	if(new_chars)
	{
		replace_mask(text, old_chars, new_chars);
		free(new_chars);
	}
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
void __stdcall replace_mask_xl12(wchar_t *text, wchar_t *old_chars, xloper12 *p_new_chars)
{
#pragma EXPORT

	cpp_xloper NewChars(p_new_chars);
	wchar_t *new_chars = NewChars.IsStr() ? (wchar_t *)NewChars : NULL;
	if(new_chars)
	{
		replace_mask(text, old_chars, new_chars);
		free(new_chars);
	}
}
//============================================================================
// Core functions are the std library byte and wide-char string functions
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
void __stdcall reverse_text_xl4(char *text)
{
#pragma EXPORT

	strrev(text);
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
void __stdcall reverse_text_xl12(wchar_t *text)
{
#pragma EXPORT

	wcsrev(text);
}
//============================================================================
// Core functions
size_t find_first(char *text, char *search_text);
size_t find_first(wchar_t *text, wchar_t *search_text);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
size_t __stdcall find_first_xl4(char *text, char *search_text)
{
#pragma EXPORT

	return find_first(text, search_text);
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
size_t  __stdcall find_first_xl12(wchar_t *text, wchar_t *search_text)
{
#pragma EXPORT

	return find_first(text, search_text);
}
//============================================================================
// Core functions
size_t find_last(char *text, char ch);
size_t find_last(wchar_t *text, wchar_t ch);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
size_t __stdcall find_last_xl4(char *text, xloper *p_ch)
{
#pragma EXPORT

	cpp_xloper Ch(p_ch);
	char ch;
	if(Ch.IsStr())
		ch = (char)Ch.First();
	else if(Ch.IsNum())
		ch = (char)(double)Ch;
	else
		return 0;

	return find_last(text, ch);
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
size_t __stdcall find_last_xl12(wchar_t *text, xloper12 *p_ch)
{
#pragma EXPORT

	cpp_xloper Ch(p_ch);
	wchar_t ch;
	if(Ch.IsStr())
		ch = Ch.First();
	else if(Ch.IsNum())
		ch = (wchar_t)(double)Ch;
	else
		return 0;

	return find_last(text, ch);
}
//============================================================================
// Core functions
int compare_text(char *a_text, char *b_text, bool case_sensitive);
int compare_text(wchar_t *a_text, wchar_t *b_text, bool case_sensitive);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
xloper * __stdcall compare_text_xl4(char *a_text, char *b_text, xloper *is_case_sensitive)
{
#pragma EXPORT

	cpp_xloper CaseSensistive(is_case_sensitive);
	bool case_sensitive = !CaseSensistive.IsFalse();
	int ret_val = compare_text(a_text, b_text, case_sensitive);
	if(ret_val == -2) // compare_text error value
		return p_xlErrValue;

	cpp_xloper RetVal(ret_val);
	return RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
xloper12 * __stdcall compare_text_xl12(wchar_t *a_text, wchar_t *b_text, xloper12 *is_case_sensitive)
{
#pragma EXPORT

	cpp_xloper CaseSensistive(is_case_sensitive);
	bool case_sensitive = !CaseSensistive.IsBool() || CaseSensistive.IsTrue();
	int ret_val = compare_text(a_text, b_text, case_sensitive);
	if(ret_val == -2) // compare_text error value
		return p_xl12ErrValue;

	cpp_xloper RetVal(ret_val);
	return RetVal.ExtractXloper12();
}
//============================================================================
// Core functions are the std library byte and wide-char string functions
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
xloper * __stdcall compare_nchars_xl4(char *a_text, char *b_text, size_t n_chars, xloper *case_sensitive)
{
#pragma EXPORT

	if(!a_text || !b_text || !n_chars || n_chars > MAX_XL4_STR_LEN)
		return p_xlErrNum;

// Case-sensitive unless explicitly Boolean False
	int ret_val = case_sensitive->xltype != xltypeBool
		|| case_sensitive->val.xbool == 1 ?
		strncmp(a_text, b_text, n_chars) :
		strnicmp(a_text, b_text, n_chars);

	cpp_xloper RetVal(ret_val);
	return RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
xloper12 * __stdcall compare_nchars_xl12(wchar_t *a_text, wchar_t *b_text, size_t n_chars, xloper12 *case_sensitive)
{
#pragma EXPORT

	if(!a_text || !b_text || !n_chars || n_chars > MAX_XL12_STR_LEN)
		return p_xl12ErrNum;

// Case-sensitive unless explicitly Boolean False
	int ret_val = case_sensitive->xltype != xltypeBool
		|| case_sensitive->val.xbool == 1 ?
		wcsncmp(a_text, b_text, n_chars) :
		wcsnicmp(a_text, b_text, n_chars);

	cpp_xloper RetVal(ret_val);
	return RetVal.ExtractXloper12();
}
//============================================================================
// Core functions
size_t find_first_excluded(char *text, char *search_text);
size_t find_first_excluded(wchar_t *text, wchar_t *search_text);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers and byte-string
size_t __stdcall find_first_excluded_xl4(char *text, char *search_text)
{
#pragma EXPORT

	return find_first_excluded(text, search_text);
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s and Unicode string
size_t __stdcall find_first_excluded_xl12(wchar_t *text, wchar_t *search_text)
{
#pragma EXPORT

	return find_first_excluded(text, search_text);
}
//============================================================================
// Core function
void ndist_inv(double prob, cpp_xloper &RetVal);
//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers
xloper * __stdcall ndist_inv_xl4(double prob)
{
#pragma EXPORT
	cpp_xloper RetVal;
	ndist_inv(prob, RetVal);
	return RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s
xloper12 * __stdcall ndist_inv_xl12(double prob)
{
#pragma EXPORT
	cpp_xloper RetVal;
	ndist_inv(prob, RetVal);
	return RetVal.ExtractXloper12();
}
//============================================================================
//============================================================================
//============================================================================
//============================================================================

//---------------------------------------------------------
// SINGLE-INTERFACE METHOD:
//
// Simple array of arguments to xlfRegister. Functions 
// that use Excel12 data types can be described here
// but xlfRegister will fail if running under Excel 11-.
//
// This method is suitable for functions that satisfy 
// one or more of these criteria:
// 1. Only use data types supported in v11- and v12+
// 2. Behaviour does not change between v11- and v12+
// 3. The performance cost of running v11- data-types
//    in v12+, or of not being registered as thread-safe
//    is not important
// 4. Are specific to v12+ and so not required in v11-
//---------------------------------------------------------
// STRUCTURE 1:
// Single interface per function regardless of Excel version.
// Data used by the function that iterates through populating
// xlopers/xloper12s, and which then calls xlfRegister.  
/*
typedef struct
{
// REQUIRED
	char *name_in_code;		// Arg2: Function name as in code (v11-)
	char *types;			// Arg3: Return type and argument types (v11-)
	char *ws_name;			// Arg4: Fn name as it appears on worksheet
	char *arg_names;		// Arg5: Argument names (Excel 11-: max 30)
// OPTIONAL:
	char *fn_category;		// Arg7: Function category for Function Wizard
	char *help_file;		// Arg9: Help file (optional)
	char *fn_description;	// Arg10: Function description text (optional)
	char *arg_help[MAX_XL12_UDF_ARGS - 11]; // Arg11...: Arg descriptions
}
	ws_func_export_data;
*/

ws_func_export_data WsFuncExports[NUM_SINGLE_IFACE_FUNCTIONS] =
{
//---------------------------------------------------------
// Chapter 2
//---------------------------------------------------------
// Excel recalulation logic examples
//---------------------------------------------------------
	{
		"C_indirect1",
		"BJJJ",
		"C_INDIRECT1",
		"Trigger,RefRow,RefCol",
		"Example",
		"",
		"Performance indirection test",
		"Trigger argument",
		"Cell ref row",
		"Cell ref col",
		"",
	},
	{
		"C_indirect2",
		"BJR",
		"C_INDIRECT2",
		"Trigger,Ref",
		"Example",
		"",
		"Performance indirection test. (Volatile)",
		"Trigger argument",
		"Cell ref",
		"",
	},
//---------------------------------------------------------
// Chapter 6
//---------------------------------------------------------
	{
		"xl_array_example1",
		"KJJ",
		"XlArrayExample1",
		"Rows,Columns",
		"Example",
		"",
		"Returns array using FP structure",
		"Rows",
		"Columns",
		"",
	},
	{
		"xl_array_example2",
		"1K",
		"XlArrayExample2",
		"InputArray",
		"Example",
		"",
		"Returns array using FP structure modified in place",
		"InputArray",
		"",
	},
	{
		"xl_array_example3",
		"1K",
		"XlArrayExample3",
		"InputArray",
		"Example",
		"",
		"Returns transposed array using FP structure modified in place",
		"InputArray",
		"",
	},
	{
		"xl_array_example4",
		"1K",
		"XlArrayExample4",
		"InputArray",
		"Example",
		"",
		"Returns array of a reduced size using FP structure modified in place",
		"InputArray",
		"",
	},
//---------------------------------------------------------
// Chapter 7
//---------------------------------------------------------
	{
		"random_array",
		"RJJ#!",
		"RandomArray",
		"Rows,Cols",
		"Example",
		"",
		"Returns array filled with random numbers",
		"Rows",
		"Cols",
		"",
	},
//---------------------------------------------------------
// Chapter 8
//---------------------------------------------------------
	{
		"evaluate",
		"RP",
		"EvaluateFormula",
		"Formula",
		"Example",
		"",
		"Evaluate using xlfEvaluate (Type 1 fn)",
		"Formula",
		"",
	},
	{
		"evaluate2",
		"RP#",
		"EvaluateFormula2",
		"Formula",
		"Example",
		"",
		"Evaluate using xlfEvaluate (Type 2 fn)",
		"Formula",
		"",
	},
	{
		"set_bin_string",
		"RCP#",
		"SetBinString",
		"Name,String",
		"Example",
		"",
		"Create (delete) a binary string in the active sheet",
		"Name",
		"(optional) string.  If not a string, name is deleted.",
		"",
	},
	{
		"get_bin_string",
		"RC#!",
		"GetBinString",
		"Name",
		"Example",
		"",
		"Get a binary name string from the active sheet",
		"Name",
		"",
	},
	{
		"function_break_example",
		"BP",
		"FnBreakExample",
		"Arg",
		"Example",
		"",
		"Function break example",
		"Arg ",
	},
	{
		"oper_type_str",
		"CP",
		"OperTypeStr",
		"Argument",
		"Example",
		"",
		"Returns type of input argument cast down to an oper",
		"Argument",
		"",
	},
	{
		"xloper_type_str",
		"CR",
		"XloperTypeStr",
		"Argument",
		"Example",
		"",
		"Returns type of the input argument. (Volatile)",
		"Argument",
		"",
	},
	{
		"get_stack",
		"B",
		"GetStack",
		"",
		"Example",
		"",
		"Returns stack space",
		"",
	},
	{
		"sheet_name",
		"RB!",
		"SheetName",
		"SheetID",
		"Example",
		"",
		"Returns the sheet name from internal ID number",
		"ID",
		"",
	},
	{
		"sheet_id",
		"RC!",
		"SheetId",
		"SheetName",
		"Example",
		"",
		"Returns the internal sheet ID number",
		"Sheet name in [Book1.xls]Sheet1.  Active sheet if omitted",
		"",
	},
	{
		"count_used_cells",
		"RJJJJ",
		"CountUsedCells",
		"FirstRow,LastRow,FirstCol,LastCol",
		"Example",
		"",
		"Counts the non-empty cells",
		"FirstRow",
		"LastRow",
		"FirstCol",
		"LastCol",
		"",
	},
	{
		"XLM4",
		"RJRRRRRRRRRRRRRRRRRRR#", // Volatile by default because of R argument and #
		"XLM4",
		"FnNum,Arg1,Arg2,Arg3,Arg4,Arg5,Arg6,Arg7,Arg8,Arg9,Arg10,Arg11,Arg12,Arg13,Arg14,Arg15,Arg16,Arg17,Arg18,Arg19",
		"Example",
		"",
		"Calls Excel4v(). (Volatile)",
		"FnNum",
		"Arg1",		"Arg2",		"Arg3",		"Arg4",
		"Arg5",		"Arg6",		"Arg7",		"Arg8",
		"Arg9",		"Arg10",	"Arg11",	"Arg12",
		"Arg13",	"Arg14",	"Arg15",	"Arg16 ",
		"Arg17",	"Arg18",	"Arg19 ",
	},
	{
		"get_caller_info",
		"RJ#!",
		"GetCallerInfo",
		"InfoParam",
		"Example",
		"",
		"Returns info about the caller cell",
		"An integer corresponding the info required",
		"",
	},
	{
		"toggle_caller",
		"R#!",
		"ToggleCaller",
		"",
		"Example",
		"",
		"Toggle the value of the calling cell ",
	},
	{
		"increment_caller",
		"RP#!",
		"IncrementCaller",
		"Trigger",
		"Example",
		"",
		"Increment the value of the calling cell by one",
		"A trigger ",
	},
	{
		"app_title",
		"RR#",  // Volatile by default because of R argument and #
		"AppTitle",
		"Title",
		"Example",
		"",
		"Change the application title. (Volatile)",
		"Title text",
		"",
	},
	{
		"get_cell",
		"RRJ#", // Volatile by default because of R argument and #
		"GetCell",
		"CellRef,InfoParam",
		"Example",
		"",
		"Returns info about the given cell. (Volatile)",
		"The cell ref in question",		
		"An integer corresponding the info required",
		"",
	},
	{
		"xl_array_transpose",
		"1K",
		"TransposeArray",
		"Input",
		"Example",
		"",
		"Transpose an array of numbers",
		"Input array ",
	},
	{
		"ref_text",
		"RRJ#",  // Volatile by default because of R argument and #
		"WsRefText", // Can't call it RefText as this conflicts with a macro sheet fn
		"Ref,A1Style",
		"Example",
		"",
		"Returns text reference. (Volatile)",
		"the reference",
		"boolean A1 = true ",
	},
	{
		"get_def",
		"RCCJ#!",
		"GetDefinition",
		"DefText,DocText,NameType",
		"Example",
		"",
		"Returns the name corresponding to the definition",
		"is the definition as text",
		"is the current book if omitted",
		"1=unhidden,2=hidden,3=all ",
	},
	{
		"name_array",
		"RCJC#!",
		"NameArray",
		"Workbook,NameType,Mask",
		"Example",
		"",
		"Returns array of names",
		"is the current book if omitted",
		"1=unhidden,2=hidden,3=all",
		"is the search mask ",
	},
	{
		"range_name",
		"RRP#",  // Volatile by default because of R argument and #
		"RangeName",
		"Name,Internal",
		"Example",
		"",
		"Returns defined name coprresponding to the given range reference. (Volatile)",
		"cell or range reference",
		"TRUE = get DLL name, FALSE, omitted = worksheet name ",
	},
	{
		"get_name",
		"RCP#!",
		"GetName",
		"Name,InfoType",
		"Example",
		"",
		"Returns name definition",
		"the range name",
		"the info type ",
	},
	{
		"check_name",
		"RC#!",
		"CheckName",
		"Name",
		"Example",
		"",
		"Returns status of defined name",
		"the range name ",
	},
	{
		"xll_name",
		"RCRPP#", // Volatile by default because of R argument and #
		"XllName",
		"Name,Definition,AsValue,Delete",
		"Example",
		"",
		"Evaluates name. (Volatile)",
		": Name of range to be defined or evaluated in DLL",
		": If omitted, evaluates previously defined name.",
		": True = Define as value. False or omitted = define as ref."
		": True = Delete. False or omitted create. "
	},
	{
		"get_formula",
		"RR#",  // Volatile by default because of R argument and #
		"GetFormula",
		"CellRef",
		"Example",
		"",
		"Returns formula. (Volatile)",
		"CellRef",
		"",
	},
//---------------------------------------------------------
// Chapter 9
//---------------------------------------------------------
// VB versus C timing functions and tests
//---------------------------------------------------------
	{
		"C_call_test",
		"BB",
		"C_call_test",
		"Arg",
		"Example",
		"",
		"Returns its argument, a number",
		"Arg",
		"",
	},
	{
		"C_timed_test0",
		"BJJJ",
		"C_TIMED_TEST0",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test0",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test1",
		"BJJJ",
		"C_TIMED_TEST1",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test1",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test2",
		"BJJJ",
		"C_TIMED_TEST2",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test2",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test3",
		"BJJJ",
		"C_TIMED_TEST3",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test3",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test4",
		"BJJJ",
		"C_TIMED_TEST4",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test4",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test5",
		"BJJJ",
		"C_TIMED_TEST5",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test5",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test6",
		"BJJJ",
		"C_TIMED_TEST6",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test6",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test7",
		"BJJJ",
		"C_TIMED_TEST7",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test7",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test8",
		"BJJJ",
		"C_TIMED_TEST8",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test8",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test9",
		"BJJJJ",
		"C_TIMED_TEST9",
		"Trigger,InnerLoops,OuterLoops,ArraySize",
		"Performance",
		"",
		"Performance test9",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"Size of array to (de)allocate"
		"",
	},
	{
		"C_timed_test10",
		"BJJJ",
		"C_TIMED_TEST10",
		"Trigger,InnerLoops,OuterLoops",
		"Performance",
		"",
		"Performance test10",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"",
	},
	{
		"C_timed_test11",
		"BJJJF",
		"C_TIMED_TEST11",
		"Trigger,InnerLoops,OuterLoops,TestString",
		"Performance",
		"",
		"Performance test11",
		"Trigger argument",
		"Number of inner loop iteration cycles to perform",
		"Number of outer loop iteration cycles to perform",
		"Test String"
		"",
	},
	{
		"get_xll_names",
		"RJ#",
		"GetXllNames",
		"Trigger",
		"Example",
		"",
		"Returns 2-column list of internal XLL names",
		"trigger ",
	},
	{
		"name_me",
		"RJ#!",
		"NameMe",
		"Create",
		"Example",
		"",
		"Creates and returns an internally defined name for this cell",
		"if true, creates a name, if false deletes an exsiting one ",
	},
	{
		"LongTaskInterfaceFn",
		"RP!#",
		"LongTask",
		"Count",
		"Example",
		"",
		"Performs long task in backgroun: counting to arg value",
		"a number > 0 ",
	},
//---------------------------------------------------------
// Chapter 10
//---------------------------------------------------------
// xloper * __stdcall SabrCalcAlpha(double Texp, double AtmVol,
//		double Fwd, double Beta, double VolVol, double Rho,
//		xloper *p_is_normal)
	{
		"SabrCalcAlpha",
		"RBBBBBBP",
		"SabrCalcAlpha",
		"Texp,AtmVol,Fwd,Beta,VolVol,Rho,IsNormal",
		"Example",
		"",
		"Calculate SABR Alpha",
		"Texp",
		"AtmVol",
		"Fwd",
		"Beta",
		"VolVol",
		"Rho",
		"IsNormal",
		"",
	},
//	xloper * __stdcall SabrCalcVol(double Texp,
//		double Fwd, double Strike, double Alpha, double Beta, 
//		double VolVol, double Rho, xloper *p_is_normal)

	{
		"SabrCalcVol",
		"RBBBBBBBP",
		"SabrCalcVol",
		"Texp,Fwd,Strike,Alpha,Beta,VolVol,Rho,IsNormal",
		"Example",
		"",
		"Calculate SABR Volatility",
		"Texp",
		"Fwd",
		"Strike",
		"Alpha",
		"Beta",
		"VolVol",
		"Rho",
		"IsNormal",
		"",
	},
// xloper * __stdcall BlackOpt(double t_exp, double fwd, double strike, double vol)
	{
		"BlackOpt",
		"RBBBB",
		"BlackOpt",
		"Texp,Fwd,Strike,Vol",
		"Example",
		"",
		"Calculate Black option. Returns 6-elt row vector",
		"Texp",
		"Fwd",
		"Strike",
		"Vol",
		"",
	},
// double __stdcall CmsPortfolioCost(double Texp, double delta_T,
//	double fwd_swap, double short_rate, double strike, int term_yrs,
//	int fixed_pmts_pa, int is_call, double delta_S, double max_strike,
//	double min_value_increment, double lambda,
//	double Alpha, double Beta, double Rho, double VolVol)
	{
		"CmsPortfolioCost",
		"BBBBBBJJJBBBBBBBB",
		"CmsPortfolioCost",
		"Texp,DeltaT,Fwd,ShortRate,Strike,TermYrs,FxdPmtsPa,IsCall,DeltaS,MaxS,MinValIncr,Lambda,Alpha,Beta,Rho,VolVol",
		"Example",
		"",
		"Calculate CMS caplet/floorlet portfolio cost",
		"Texp",
		"Length of caplet (years)",
		"Fwd swap rate",
		"Fwd caplet-period short rate",
		"Strike",
		"Term of the underlying swap",
		"Swap fixed rate payments per annum",
		"True or non-zero if caplet, else false or zero",
		"Strike step for portfolio swaptions",
		"Maximum strike (for caplets)",
		"Minimum value increment for additions to portfolio",
		"Fraction of strike step above current strike for fixing rate valuation",
		"SABR parameter alpha: volatility of the underlying",
		"SABR parameter beta: normal (0) to lognormal (1)",
		"SABR parameter rho: (-1 to 1) correlation of underlying and vol processes",
		"SABR parameter vol of vol: lognormal vol of vol"
	},
//---------------------------------------------------------------
// Miscellaneous example functions
//---------------------------------------------------------------
//	xloper * __stdcall no_error(xloper *p_input, xloper *pUDF)
	{
		"no_error",
		"RPP#",
		"NoErr",
		"Input,UdfTestFn",
		"Example",
		"",
		"Returns the current values unless an error.",
		"the value(s) or cell(s) to be tested and returned if not an error",
		"(optional) UDF that returns TRUE if passed an error value  ",
	},
// xloper * __stdcall ThreadSafeFunction(xloper *pTrigger, xloper *pLoopCount)
	{
		"ThreadSafeFunction",
		"RPP$",
		"ThreadSafeFunction",
		"Trigger,LoopCount",
		"Example",
		"",
		"MT calc test function (thread-safe). Returns ThreadID, Start, End, OpAdr",
		"trigger",
		"loop counter (to regulate execution time)  ",
	},
// xloper * __stdcall ThreadUnsafeFunction(xloper *pTrigger, xloper *pLoopCount)
	{
		"ThreadUnsafeFunction",
		"RPP",
		"ThreadUnsafeFunction",
		"Trigger,LoopCount",
		"Example",
		"",
		"MT calc test function (not thread-safe). Returns ThreadID, Start, End, OpAdr",
		"trigger",
		"loop counter (to regulate execution time)  ",
	},
// xloper12 * __stdcall ThreadSafeFunction12(xloper12 *pTrigger, xloper12 *pLoopCount)
	{
		"ThreadSafeFunction12",
		"UQQ$",
		"ThreadSafeFunction12",
		"Trigger,LoopCount",
		"Example",
		"",
		"MT calc test function (thread-safe). Returns ThreadID, Start, End, OpAdr",
		"trigger",
		"loop counter (to regulate execution time)  ",
	},
// xloper12 * __stdcall ThreadUnsafeFunction12(xloper12 *pTrigger, xloper12 *pLoopCount)
	{
		"ThreadUnsafeFunction12",
		"UQQ",
		"ThreadUnsafeFunction12",
		"Trigger,LoopCount",
		"Example",
		"",
		"MT calc test function (not thread-safe). Returns ThreadID, Start, End, OpAdr",
		"trigger",
		"loop counter (to regulate execution time)  ",
	},
// xl12_array * __stdcall xl12_array_example1(INT32 rows, INT32 columns)
	{
		"xl12_array_example1",
		"K%JJ",
		"Xl12ArrayExample1",
		"Rows,Columns",
		"Example",
		"",
		"Returns array using FP structure",
		"Rows",
		"Columns",
		"",
	},
};
//============================================================================
//============================================================================
//============================================================================
FnRegData RT_name_test( // Instance name is not important
	"name_test",	// Function name as in code (v11-)
	"RPP#",			// Return type and argument types (v11-)
	"",				// Function name as in code (v12+)
	"",				// Return type and argument types (v12+)
	"NameTest",		// Fn name as it appears on worksheet
	"Name,Arg",		// Argument names (Excel 11-: max 30)
	"",				// Argument names (Excel 12+: max 255)
	"Example",		// Function category for Function Wizard
	"",				// Help file (optional)
	"Evaluates a name", // Function description text (optional)
	"the name",		// Fn arg help text strings...
	"arg  ",
	NULL); // Arg list terminator


xloper * __stdcall name_test(xloper *pName, xloper *pGetTestNameDefn)
{
#pragma EXPORT
	cpp_xloper Arg(pGetTestNameDefn);

	if(Arg == 0)
	{
		xlName DllName;
		cpp_xloper Defn((RW)0, (RW)0, (COL)0, (COL)0); // A1
		DllName.Define("DllTestName", Defn, true);
		return p_xlTrue;
	}
	else if(Arg == 1)
	{
	// Return the definition
		cpp_xloper Name("DllTestName");
		Name.Excel(xlfGetName, 1, &Name);
		return Name.ExtractXloper();
	}
	else if(Arg == 2)
	{
		cpp_xloper RetVal;
		RetVal.Excel(xlfEvaluate, 1, pName);
		RetVal = (int)RetVal.GetType();
		return RetVal.ExtractXloper();
	}
	else if(Arg == 4)
	{
		cpp_xloper RetVal;
		RetVal.Excel(xlfEvaluate, 1, pName);
		return RetVal.ExtractXloper();
	}
	return p_xlErrValue;
}
//============================================================================
// Example UDF function returning true if p_op is an error value or a string
// starting with '#'.

FnRegData RT_is_error_UDF( // Instance name is not important
	"is_error_UDF",	// Function name as in code (v11-)
	"RP",			// Return type and argument types (v11-)
	"",				// Function name as in code (v12+)
	"",				// Return type and argument types (v12+)
	"IsErrUdf",		// Fn name as it appears on worksheet
	"Input",		// Argument names (Excel 11-: max 30)
	"",				// Argument names (Excel 12+: max 255)
	"Example",		// Function category for Function Wizard
	"",				// Help file (optional)
	"Example UDF function.  Returns TRUE if passed an error value or string starting with #.", // Function description text (optional)
	"the value or single cell ref to be tested",		// Fn arg help text strings...
	NULL); // Arg list terminator


xloper * __stdcall is_error_UDF(xloper *p_op)
{
#pragma EXPORT
	cpp_xloper Op(p_op);

// If a reference type, coerce to a value (dereference) before testing
// Don't need this if p_op registered as type P
	if(Op.IsRef() && !Op.ConvertRefToSingleValue())
		return p_xlTrue; // Could not coerce

	if(Op.IsErr() || (Op.IsStr() && Op.First() == L'#'))
		return p_xlTrue;

	return p_xlFalse;
}
//============================================================================
// Can only call xlfGetWorkspace from a macro-sheet function or command, so if
// that fails, then is a worksheet function
bool caller_is_ws_function(void)
{
// Retrieve the R1C1 state. Fails if called from WS function
// Don't need the return value so pass NULL.
	if(gExcelVersion12plus)
	{
		xloper12 four;
		four.val.w = 4; // Ask if Excel is in R1C1 mode
		four.xltype = xltypeInt;
		return Excel12(xlfGetWorkspace, 0, 1, &four) != xlretSuccess;
	}
	else
	{
		xloper four;
		four.val.w = 4; // Ask if Excel is in R1C1 mode
		four.xltype = xltypeInt;
		return Excel4(xlfGetWorkspace, 0, 1, &four) != xlretSuccess;
	}
}
//============================================================================
bool is_error(xloper *p_op, xloper *pUDF = NULL)
{
	if(pUDF && pUDF->xltype == xltypeStr)
	{
		cpp_xloper UdfRetVal;
		return (UdfRetVal.Excel(xlUDF, 2, pUDF, p_op) != xlretSuccess)
			|| UdfRetVal.IsTrue();
	}
// Default test if UDF is not a string
	return p_op->xltype == xltypeErr;
}
//============================================================================
xloper * __stdcall no_error(xloper *p_input, xloper *pUDF)
{
#pragma EXPORT

	cpp_xloper Caller;

	if(p_input->xltype == xltypeMulti) // need to check elt-by-elt
	{
		Caller.Excel(xlfCaller);

		if(!Caller.ConvertRefToMulti())
			return p_xlErrNa;

// First check that input and calling range are same size and shape
		cpp_xloper Input(p_input); // Shallow copy
		if(!Input.SameShapeAs(Caller))
			return p_xlErrValue;

		DWORD i, size;
		Input.GetArraySize(size);
		xloper *p_op = p_input->val.array.lparray;

		for(i = 0; i < size; i++, p_op++)
		{
			if(is_error(p_op, pUDF))
				continue;

			// else copy the input value into the array
			Caller.SetArrayElt(i, p_op);
		}
		return Caller.ExtractXloper();
	}
	if(is_error(p_input, pUDF))
		goto rtn_current_val;

	return p_input;

rtn_current_val:
	Caller.Excel(xlfCaller);
	if(!Caller.IsType(xltypeSRef | xltypeRef))
		return NULL;

	cpp_xloper RetVal;
	RetVal.Excel(xlCoerce, 1, &Caller);
	if(RetVal.IsType(xltypeErr))
		RetVal = 0.0;
	return RetVal.ExtractXloper();
}

char *string_lookup_function(xloper *p_op)
{
	return "test string";
}

xloper * __stdcall thread_unsafe_example(xloper *pArg)
{
#pragma EXPORT

	static xloper ret_val;

	char *str = string_lookup_function(pArg);
	BYTE str_len = strlen(str); // <= 255
	ret_val.val.str = (char *)malloc(str_len + 2); // len + string + null terminator
	ret_val.val.str[0] = str_len;
	strcpy(ret_val.val.str + 1, str);
	ret_val.xltype = xltypeStr | xlbitDLLFree;

	return &ret_val;
}
