//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
#pragma once

#include "cpp_xloper.h"
// Thread-local storage function prototypes and defintitions
#include "TLS.h"

#if _MSC_VER > 1200 // Later than Visual Studio 6.0
#define EXPORT comment(linker, "/EXPORT:"__FUNCTION__"="__FUNCDNAME__)
#else // else need to use DEF file or __declspec(dllexport)
#define EXPORT
#endif

#define NUM_COMMANDS					13
#define NUM_SINGLE_IFACE_FUNCTIONS		59
#define NUM_DUAL_IFACE_FUNCTIONS		24

extern xloper fn_ID[NUM_SINGLE_IFACE_FUNCTIONS];
extern xloper dual_fn_ID[NUM_DUAL_IFACE_FUNCTIONS];
extern xloper cmd_ID[NUM_COMMANDS];


// These are displayed by the Excel add-in manager
extern char *AddinVersionStr;
extern char *AddinName;

extern int gExcelVersion; // default is 0: version not known
extern bool gExcelVersion12plus; // default is false
extern bool gExcelVersion11minus; // default is true

bool caller_is_ws_function(void);

// Enumeration of the columns of a custom C API dialog box

enum
{
	CAPI_DLG_ITEM,
	CAPI_DLG_HPOS,
	CAPI_DLG_VPOS,
	CAPI_DLG_WIDTH,
	CAPI_DLG_HEIGHT,
	CAPI_DLG_TEXT,
	CAPI_DLG_VALUE,
	CAPI_DLG_COLUMNS,
};

bool called_from_paste_fn_dlg(void);
bool called_from_replace_dlg(void);
bool called_from_Excel_dlg(void);

#define MAX_DOUBLE		(1.7976931348623158e308)

// Argument structures for xlfRegister function.

// Arg'n' denotes the number of the argument passed to xlfRegister.
// Arg1 is the DLL path and file name. It is the same for all exports
// and is populated elsewhere.  Arg6 is always 1 (or omitted) for
// worksheet functions and so on not required in this structure

// STRUCTURE 1:
// Single interface per function regardless of Excel version.
// Data used by the function that iterates through populating
// xlopers/xloper12s, and which then calls xlfRegister.  

// xlfRegister arguments not included in ws_func_export_data
// Arg1: DLL function path and filename
// Arg6: macro_type - always 1 for worksheet functions, so omitted here
// Arg8: Shortcut key (Mac only) - omitted

typedef struct
{
// REQUIRED (see comments above):
	char *name_in_code;		// Arg2: Function name as in code (v11-)
	char *types;			// Arg3: Return type and argument types (v11-)
	char *ws_name;			// Arg4: Fn name as it appears on worksheet
	char *arg_names;		// Arg5: Argument names (Excel 11-: max 30)
// OPTIONAL:
	char *fn_category;		// Arg7: Function category for Function Wizard
	char *help_file;		// Arg9: Help file (optional)
	char *fn_description;	// Arg10: Function description text (optional)
	char *arg_help[MAX_XL12_UDF_ARGS - 11]; // Arg11...: Arg descriptions
}
	ws_func_export_data;


// STRUCTURE 2:
// Dual interface per function: one for v11- and one for v12+.
// Data used by the function that iterates through populating
// xlopers/xloper12s, and which then calls xlfRegister.  

// If name_in_code12 is omitted, types12 and arg_names12 are ignored
// and (even when running with Excel 12+) the Excel 11- interface is
// registered.

// If name_in_code12 is provided and running with Excel 12+ types12
// is required.  If arg_names12 is not supplied, arg_names is used.
// The Excel 11- interface is not registered.

// 11 args preceed 1st arg_help in calls to xlfRegister, so at most can only
// supply MAX_XL12_UDF_ARGS - 11 help text strings in Excel 12.  If running
// with Excel 11- only the first MAX_XL11_UDF_ARGS - 11 of these will be used

// xlfRegister arguments not included in dual_ws_func_export_data
// Arg1: DLL function path and filename
// Arg6: macro_type - always 1 for worksheet functions, so omitted here
// Arg8: Shortcut key (Mac only) - omitted

typedef struct
{
// REQUIRED (see comments above):
	char *name_in_code;		// Arg2: Function name as in code (v11-)
	char *types;			// Arg3: Return type and argument types (v11-)
	char *name_in_code12;	// Arg2: Function name as in code (v12+)
	char *types12;			// Arg3: Return type and argument types (v12+)
	char *ws_name;			// Arg4: Fn name as it appears on worksheet
	char *arg_names;		// Arg5: Argument names (Excel 11-: max 30)
	char *arg_names12;		// Arg5: Argument names (Excel 12+: max 255)
// OPTIONAL:
	char *fn_category;		// Arg7: Function category for Function Wizard
	char *help_file;		// Arg9: Help file (optional)
	char *fn_description;	// Arg10: Function description text (optional)
	char *arg_help[MAX_XL12_UDF_ARGS - 11]; // Arg11...: Fn arg help text strings
}
	dual_ws_func_export_data;

// Used to register DLL functions and commands
extern char *CommandExports[NUM_COMMANDS][2];
extern ws_func_export_data WsFuncExports[NUM_SINGLE_IFACE_FUNCTIONS];
extern dual_ws_func_export_data DualWsFuncExports[NUM_DUAL_IFACE_FUNCTIONS];

xloper register_function(ws_func_export_data *ps);
bool register_dual_function(dual_ws_func_export_data *ps, cpp_xloper &RetVal);
xloper register_command(char *code_name, char *Excel_name);
bool unregister_function(int fn_index);
bool unregister_dual_function(int fn_index);
bool unregister_command(int cmd_index);


// STRUCTURE 3:
// Class can be instantiated near, but outside, function code.
// Argument classes for xlfRegister function. Used to register
// DLL functions (class FnRegData) and commands (class CmdRegData)

// ---------------------------------------------------------------
// ---------------------- class FnRegData ------------------------
// ---------------------------------------------------------------

/*

Class supports a dual interface per function: one for v11- and one
for v12+.  One of these only is registered depending on the Excel
version that is running.

If m_NameInCode12 is omitted and Excel 12+ is running then:
	- m_Types12 and m_ArgNames12 are ignored
	- the Excel 11- interface is registered

If m_NameInCode12 is provided and Excel 12+ is running then:
	- m_Types12 is used if given, otherwise m_Types is used.
	- m_ArgNames12 is used if given, otherwise m_ArgNames is used.
	- the Excel 11- interface function is not registered.

11 args precede the 1st argument help string, m_ArgHelp[0], in calls
to xlfRegister, so at most can only handle MAX_XL12_UDF_ARGS - 11
help text strings in Excel 12+. If running with Excel 11- only
the first MAX_XL11_UDF_ARGS - 11 of these will be used.

xlfRegister arguments not supplied to the class' constructor are:
Arg1: DLL function path and filename
Arg6: macro_type - always 1 for worksheet functions, so omitted here
Arg8: Shortcut key (Mac only) - omitted

Arg'n' denotes the number of the argument passed to xlfRegister
counting from 0. Arg1 is the DLL path and file name. It is the
same for all exports and is populated elsewhere.  Arg6 is always
1 (or omitted) for worksheet functions and so on not required in
this structure

*/

// Class can be instantiated near, but outside, function code.
class FnRegData
{
public:
	FnRegData(void) {return;}
	FnRegData(char *name_in_code, char *types, char *name_in_code12,
		char *types12, char *ws_name, char *arg_names, char *arg_names12,
		char *fn_category, char *help_file, char *fn_description, ...);

	static void ClearData(void)
	{
		if(InstanceArray)
		{
			delete[] InstanceArray;
			InstanceArray = NULL;
			count = array_size = 0;
		}
	}
	static void RegisterAll(void);
	bool RegisterThis(void);

private:
// Class member variables
	static long count;
	static long array_size;
	static FnRegData **InstanceArray;
// Instance member variables
	char *m_NameInCode;	// Arg2: Function name as in code (v11-)
	char *m_Types;			// Arg3: Return type and argument types (v11-)
	char *m_NameInCode12;	// Arg2: Function name as in code (v12+)
	char *m_Types12;		// Arg3: Return type and argument types (v12+)
	char *m_WsName;		// Arg4: Fn name as it appears on worksheet
	char *m_ArgNames;		// Arg5: Argument names (Excel 11-: max 30)
	char *m_ArgNames12;	// Arg5: Argument names (Excel 12+: max 255)
	char *m_FnCategory;	// Arg7: Function category for Function Wizard
	char *m_HelpFile;		// Arg9: Help file (optional)
	char *m_FnDescription;	// Arg10: Function description text (optional)
	char *m_ArgHelp[MAX_XL12_UDF_ARGS - 11]; // Arg11... arg help text
};

// ---------------------------------------------------------------
// ---------------------- class CmdRegData -----------------------
// ---------------------------------------------------------------

/*

No need to have a dual interface as commands do not take or return
version-specific data types and are always run on the main thread,
and are not subject to multi-threading constraints.

*/

// Class can be instantiated near, but outside, function code.
class CmdRegData
{
public:
	CmdRegData(void) {return;}
	CmdRegData(char *name_in_code, char *macro_name);

	static void ClearData(void)
	{
		if(InstanceArray)
		{
			delete[] InstanceArray;
			InstanceArray = NULL;
			count = array_size = 0;
		}
	}
	static void RegisterAll(void);
	bool RegisterThis(void);

private:
// Class member variables
	static long count;
	static long array_size;
	static CmdRegData **InstanceArray;
// Instance member variables
	char *m_NameInCode;		// Arg2: Function name as in code (v11-)
	char *m_MacroName;		// Arg4: Fn name as Excel knows it
};


