//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
#pragma once

#include "cpp_xloper.h"

//-------------------------------------------------------------------
//-------------------------------------------------------------------
// C++ class that encapsulates Excel named range handling using the
// C API
//-------------------------------------------------------------------
//-------------------------------------------------------------------

class xlName
{
public:
//-------------------------------------------------------------------
// constructors & destructor
//-------------------------------------------------------------------
	xlName():m_Defined(false),m_RefValid(false),m_Worksheet(false){}
	xlName(const char *name) {Set(name);} // Reference to existing range
	~xlName() {Clear();}

// Copy constructor uses operator= function
	xlName(const xlName & source) {*this = source;}

//-------------------------------------------------------------------
// Overloaded operators
//-------------------------------------------------------------------
// Object assignment operator
	xlName & operator=(const xlName& source);

//-------------------------------------------------------------------
// Assignment operators place values in cell(s) that range refers to.
// Cast operators retrieve values or assign nil if range is not valid
// or conversion was not possible.  Casting to char * will return
// dynamically allocated memory that the caller must free.
//-------------------------------------------------------------------

	int operator=(int);
	bool operator=(bool b);
	double operator=(double);
	WORD operator=(WORD e);
	const char * operator=(const char *);
	const xloper * operator=(const xloper *);
	const xloper12 * operator=(const xloper12 *);
	const cpp_xloper & operator=(const cpp_xloper &);
	const VARIANT * operator=(const VARIANT *);
	const xl4_array * operator=(const xl4_array *array);
	double operator+=(double);
	double operator++(void) {return operator+=(1.0);}
	double operator--(void) {return operator+=(-1.0);}
	operator int(void) const;
	operator bool(void) const;
	operator double(void) const;
	operator char *(void) const; // DLL-allocated copy, caller must free
	operator wchar_t *(void) const; // DLL-allocated copy, caller must free

	bool IsDefined(void) const {return m_Defined;}
	bool IsRefValid(void) const {return m_RefValid;}
	bool IsWorksheetName(void) const {return m_Worksheet;}
	char *GetDef(void) const; // get defintion as string (caller must free)
	char *GetName(void) const; // returns deep copy that caller must free
	void GetName(cpp_xloper &Name) const; // Initialises cpp_xloper to name
	bool GetRangeSize(DWORD &size) const;
	bool GetRangeSize(RW &rows, COL &columns) const;
	bool GetValues(cpp_xloper &Ref) const; // range contents as xltypeMulti
	bool GetValues(double *array, DWORD array_size) const;
	bool GetRef(cpp_xloper &Values) const; // range as xltypeRef
	bool SetValues(const cpp_xloper &Values);
	bool SetValues(const double *array, RW rows, COL columns);
	bool NameIs(const char *name) const;
	bool RefreshRef(void); // refreshes state of name and defn ref
// SetToRef sets instance to ref's name if it exists
	bool SetToRef(const cpp_xloper &Ref, bool internal);
	bool SetToCallersName(void); // set to caller's name if it exists
	bool NameCaller(const char *name); // create internal name for caller
	bool Set(const char *name); // Create a reference to an existing range
	bool Define(const cpp_xloper &Defintion, bool in_dll);
	bool Define(const char *name, const cpp_xloper &Defintion, bool in_dll);
	void Delete(void); // Delete name and free instance resources
	void Clear(void); // Clear instance memory but don't delete name
	void SetNote(const char *text); // Doesn't work - might be C API bug
	char *GetNote(void);

	static int GetNumInternalNames(void) {return m_NumInternalNames;}
	static void IncrNumInternalNames(void) {m_NumInternalNames++;}
	static void DecrNumInternalNames(void) {m_NumInternalNames--;}

private:
	static int m_NumInternalNames; // Keep a count of all created names

protected:
	bool m_Defined; // Name has been defined
	bool m_RefValid; // Name's defintion (if a ref) is valid
	bool m_Worksheet; // Name is worksheet name, not internal to DLL
	cpp_xloper m_RangeRef;
	cpp_xloper m_RangeName;
};

//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Internal name list management functions

bool add_name_record(void *fn_ptr, xlName &Range);
int __stdcall delete_all_xll_names(void);
void *find_name_record_fn_ptr(xlName &Range);
int __stdcall clean_xll_name_list(void);

//-------------------------------------------------------------------
//-------------------------------------------------------------------
