/*

IMPORTANT NOTICE.  PLEASE READ CAREFULLY

The sample code (and modifications thereto) contained in the Book and
CD are for illustrative purposes only.  

Each user must obtain a separate license from Microsoft to use the
sample code (subject to the terms of such license) when such code
is released by Microsoft. When released, the sample code (and
accompanying license) can be obtained at
http://msdn2.microsoft.com/en-us/office/aa905351.aspx.

The sample code described in the Book and CD may not be the actual
sample code that Microsoft releases.

Microsoft is not affiliated in any manner with the author or publisher
of the Book and CD and does not endorse the Book, CD or any
modifications to the sample code.

Eigensys and Microsoft disclaim any warranty in relation to the sample
code (and any modifications thereto) contained in the Book and CD.
The sample code (and any modifications thereto) are provided "as is."
You bear the risk of using it.  Eigensys and Microsoft give no express
warranties, guarantees or conditions.  You may have additional consumer
rights under your local laws which this agreement cannot change.  To
the extent permitted under your local laws, Eigensys and Microsoft
exclude the implied warranties of merchantability, fitness for a
particular purpose and non-infringement.
	
*/


// Modified by Steve Dalton.
// Definitions of Excel12 and Excel12v are changed to use
// the const specifier, to be consistent with the modified
// xlcall.h, which enables the cpp_xloper class to use 
// 'const' in the member functions wrapping calls to the
// C API.

// Modifications are contained within USE_MODIFIED_XLCALLH_DEFINITION
// preprocessor #if...#else...#endif blocks

/*
**  Microsoft Excel Developer's Toolkit
**  Version 12.0
**
**  File:           INCLUDE\XLCALL.CPP
**  Description:    Code file for Excel 2007 callbacks
**  Platform:       Microsoft Windows
**
**  This file defines the entry points 
**  which are used in the Microsoft Excel 2007 C API.
**
*/

#ifndef _WINDOWS_
#include <windows.h>
#endif

// Modified header file defines USE_MODIFIED_XLCALLH_DEFINITION = 1
#include "xlcall.h"

/*
** Excel 12 entry points backwards compatible with Excel 11
**
** Excel12 and Excel12v ensure backwards compatibility with Excel 11
** and earlier versions. These functions will return xlretFailed when
** used to callback into Excel 11 and earlier versions
*/

#define cxloper12Max 255
#define EXCEL12ENTRYPT "MdCallBack12"

#if USE_MODIFIED_XLCALLH_DEFINITION
// Modified definition of the function argument rgpxloper12 using const for compatibility
// with cpp_xloper class C API wrapper functions
typedef int (PASCAL *EXCEL12PROC)(int xlfn, int coper, const xloper12 **rgpxloper12, xloper12 *xloper12Res);
#else // Original Microsoft Excel 2007 SDK definition
typedef int (PASCAL *EXCEL12PROC) (int xlfn, int coper, LPXLOPER12 *rgpxloper12, LPXLOPER12 xloper12Res);
#endif

HMODULE hmodule;
EXCEL12PROC pexcel12;

__forceinline void FetchExcel12EntryPt(void)
{
	if (pexcel12 == NULL)
	{
		hmodule = GetModuleHandle(NULL);
		if (hmodule != NULL)
		{
			pexcel12 = (EXCEL12PROC) GetProcAddress(hmodule, EXCEL12ENTRYPT);
		}
	}
}

int _cdecl Excel12(int xlfn, LPXLOPER12 operRes, int count, ...)
{

#ifdef _WIN64
	
	return(xlretFailed);

#else

#if USE_MODIFIED_XLCALLH_DEFINITION
// Modified definition of rgxloper12 using const for compatibility
// with cpp_xloper class C API wrapper functions
	const xloper12 *rgxloper12[cxloper12Max];
#else // Original Microsoft Excel 2007 SDK definition
	LPXLOPER12 rgxloper12[cxloper12Max];
#endif
	va_list ap;
	int ioper;
	int mdRet;

	FetchExcel12EntryPt();
	if (pexcel12 == NULL)
	{
		mdRet = xlretFailed;
	}
	else
	{
		mdRet = xlretInvCount;
		if ((count >= 0)  && (count <= cxloper12Max))
		{
			va_start(ap, count);
			for (ioper = 0; ioper < count ; ioper++)
			{
				rgxloper12[ioper] = va_arg(ap, LPXLOPER12);
			}
			va_end(ap);
			mdRet = (pexcel12)(xlfn, count, &rgxloper12[0], operRes);
		}
	}
	return(mdRet);
	
#endif
}


#if USE_MODIFIED_XLCALLH_DEFINITION
// Modified definition of the function argument rgpxloper12 using const for compatibility
// with cpp_xloper class C API wrapper functions
int pascal Excel12v(int xlfn, xloper12 *operRes, int count, const xloper12 *opers[])
#else // Original Microsoft Excel 2007 SDK definition
int pascal Excel12v(int xlfn, LPXLOPER12 operRes, int count, LPXLOPER12 opers[])
#endif
{
#ifdef _WIN64
	
	return(xlretFailed);

#else

	int mdRet;

	FetchExcel12EntryPt();
	if (pexcel12 == NULL)
	{
		mdRet = xlretFailed;
	}
	else
	{
		mdRet = (pexcel12)(xlfn, count, &opers[0], operRes);
	}
	return(mdRet);

#endif
}
