#include <windows.h>
#include "xlcall.h"

#define COMPILE_DOC_EXAMPLES 0

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#if 1

#include "xlladdin.h"

FnRegData RF_IntegerTestFn( // Instance name is not important
	"TestFn_xl4",			// Function name as in code (v11-)
	"1K",					// Return type and argument types (v11-)
	"TestFn_xl12",			// Function name as in code (v12+)
	"UI$",					// Return type and argument types (v12+)
	"TestFn",				// Fn name as it appears on worksheet
	"Inupt",				// Argument names (Excel 11-: max 30)
	"",						// Argument names (Excel 12+: max 255)
	"Example",				// Function category for Function Wizard
	"",						// Help file (optional)
	"Returns input",		// Function description text (optional)
	"the inputs  ",			// Fn arg help text strings...
	NULL); // Arg list terminator

//----------------------------------------------------------------------------
// Excel 11- interface function.
void __stdcall TestFn_xl4(FP *p)
{
#pragma EXPORT
	xloper x;
	Excel4(xlfCaller, &x, 0);
	Excel4(xlFree, 0, 1, &x);
//	cpp_xloper RetVal((double)p->rows);
	return; // RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s
xloper12 * __stdcall TestFn_xl12(short i)
{
#pragma EXPORT
	cpp_xloper RetVal(i);
	return RetVal.ExtractXloper12();
}
#endif

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#if COMPILE_DOC_EXAMPLES
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------


#define MAX_SIZE_T_FP_ELTS 536870910 /* =INT((2^32 - 1)/8) - 1 */

size_t get_FP_mem_size(RW rows, COL columns)
{
// Excel 2007 has 2^14 columns x 2^20 rows, so INT32 is too small
	INT64 size64 = rows * columns; // RW and COL are INT32

// size_t based memory management functions cannot provide 
// memory blocks big enough for > MAX_SIZE_T_FP_ELTS
	if(size64 <= 0 || size64 > MAX_SIZE_T_FP_ELTS)
		return 0;

// Can safely cast size64 down to a size_t now
	return 2 * sizeof(INT32) + (size_t)size64 * sizeof(double);
}

FP *new_FP(unsigned short n_rows, unsigned short n_columns)
{
	if(n_rows <= 0 || n_columns <= 0)
		return NULL;
	size_t n = n_rows * n_columns - 1; /* number of extra elts */
	FP *p_array = (FP *)malloc(sizeof(FP) + n * sizeof(double));
	if(p_array)
	{
		p_array->rows = n_rows;
		p_array->columns = n_columns;
	}
	return p_array;
}

double get_FP12_element(FP12 *pfp, INT32 row, INT32 column)
{
	if(row < 0 || row >= pfp->rows
	|| column < 0 || column >= pfp->columns)
		return 0.0; /* would be better to throw or return an error */

	return pfp->array[row * pfp->columns + column];
}
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
bool delete_my_backup_files(bool show_dialog)
{
	XLOPER xResult, xFilter;
	xFilter.xltype = xltypeStr;
	xFilter.val.str = "\014my_data*.bak"; // String length: 14 octal
	int cmd = show_dialog ? xlcFileDelete | xlPrompt : xlcFileDelete;

// xResult should be Boolean TRUE if successful in which
// case return true, otherwise false.
	return (Excel4(cmd, &xResult, 1, &xFilter) == xlretSuccess
		&& xResult.xltype == xltypeBool
#ifdef __cplusplus
		&& xResult.val.xbool == 1);
#else
		&& xResult.val.bool == 1);
#endif
}


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
XLOPER *TempActiveRef(int a, int b, int c, int d)
{
	return NULL;
}

int WINAPI InternationlExample(void)
{
	XLOPER xResult, xSum;
	xSum.xltype = xltypeStr;
	xSum.val.str = "\015=SUM(X1:X100)";
	Excel4(xlcFormula | xlIntl, &xResult, 2, &xSum, TempActiveRef(2, 2, 1, 1));
	return 1;
}
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
void GetName_example(void)
{
	XLOPER xDllName;
	if(Excel4(xlfGetName, &xDllName, 0) == xlretSuccess)
	{
		// Use the name....
		// Then free the memory that Excel allocated for the string
		Excel4(xlFree, 0, 1, &xDllName);
	}
}

void GetName_example12(void)
{
	XLOPER12 xDllName;
	if(Excel12(xlfGetName, &xDllName, 0) == xlretSuccess)
	{
		// Use the name....
		// Then free the memory that Excel allocated for the string
		Excel12(xlFree, 0, 1, &xDllName);
	}
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void Excel12v_example(double *dbl_array, int size, double &sum, double &average, double &min, double &max)
{
// 30 is the limit in Excel 2003. 255 is the limit in Excel 2007.
// Use the lower limit to be safe, although it is better to make
// the function version-aware and use the correct limit.
	if(size < 1 || size > 30)
		return;

// Create an array of XLOPER12s
	XLOPER12 *xOpArray = new XLOPER12[size];
// Create an array of pointers to XLOPER12s
	LPXLOPER12 *xPtrArray = new LPXLOPER12[size];

// Initialise and populate the array of XLOPER12s
// and set up the pointers in the pointer array
	for(int i = 0; i < size; i++)
	{
		xOpArray[i].xltype = xltypeNum;
		xOpArray[i].val.num = dbl_array[i];
		xPtrArray[i] = xOpArray + i;
	}

	XLOPER12 xResult;
	int retval;
	int fn[4] = {xlfSum, xlfAverage, xlfMin, xlfMax};
	double *result_ptr[4] = {&sum, &average, &min, &max};

	for(i = 0; i < 4; i++)
	{
		retval = Excel12v(fn[i], &xResult, size, xPtrArray);
		if(retval == xlretSuccess && xResult.xltype == xltypeNum)
			*result_ptr[i] = xResult.val.num;
	}
	delete[] xPtrArray;
	delete[] xOpArray;
}


void Excel12_example(double *dbl_array, int size, double &sum, double &average, double &min, double &max)
{
// In this implementation the upper limit is the largest
// sinlge column array = 2^20 = 1048576 rows in Excel 2007
	if(size < 1 || size > 1048576)
		return;

// Create an array of XLOPER12s
	XLOPER12 *xOpArray = new XLOPER12[size];

// Create and initialise an xltypeMulti array
// representing a one-column array.
	XLOPER12 xOpMulti;
	xOpMulti.xltype = xltypeMulti;
	xOpMulti.val.array.lparray = xOpArray;
	xOpMulti.val.array.columns = 1;
	xOpMulti.val.array.rows = size;

// Initialise and populate the array of XLOPER12s
	for(int i = 0; i < size; i++)
	{
		xOpArray[i].xltype = xltypeNum;
		xOpArray[i].val.num = dbl_array[i];
	}

	XLOPER12 xResult;
	int fn[4] = {xlfSum, xlfAverage, xlfMin, xlfMax};
	double *result_ptr[4] = {&sum, &average, &min, &max};

	for(i = 0; i < 4; i++)
	{
		Excel12(fn[i], &xResult, 1, &xOpMulti);
		if(xResult.xltype == xltypeNum)
			*result_ptr[i] = xResult.val.num;
	}
	delete[] xOpArray;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

int WINAPI show_DLL_name(void)
{
	XLOPER12 xDllName;
	if(Excel12(xlfGetName, &xDllName, 0) == xlretSuccess)
	{
		// Display the name
		Excel12(xlcAlert, 0, 1, &xDllName);
		// Then free the memory that Excel allocated for the string
		Excel12(xlFree, 0, 1, &xDllName);
	}
	return 1;
}

LPXLOPER12 WINAPI get_DLL_name(LPXLOPER12 pxArg)
{
	static XLOPER12 xRtnValue; // Not thread-safe

	if(pxArg->xltype != xltypeBool || pxArg->val.xbool != 1)
	{
// If argument not Boolean TRUE return #N/A
		xRtnValue.xltype = xltypeErr;
		xRtnValue.val.err = xlerrNA;
		return &xRtnValue;
	}

	Excel12(xlfGetName, &xRtnValue, 0);

// If xlfGetName failed, xRtnValue will be #VALUE!
	if(xRtnValue.xltype == xltypeStr)
	{
// Tell Excel to free the string memory after
// it has copied out the return value.
		xRtnValue.xltype |= xlbitXLFree;
	}
	return &xRtnValue;
}

LPXLOPER12 WINAPI get_DLL_name_2(LPXLOPER12 pxArg)
{
	static XLOPER12 xRtnValue; // Not thread-safe

	if(pxArg->xltype != xltypeBool || pxArg->val.xbool != 1)
	{
// If argument not Boolean TRUE return #N/A
		xRtnValue.xltype = xltypeErr;
		xRtnValue.val.err = xlerrNA;
		return &xRtnValue;
	}

	Excel12(xlfGetName, &xRtnValue, 0);

// If xlfGetName failed, xRtnValue will be #VALUE!
	if(xRtnValue.xltype == xltypeStr)
		return &xRtnValue;

// Make a copy of the DLL path and filename
	wchar_t *leader = L"The full pathname for this DLL is ";
	size_t leader_len = wcslen(leader);
	size_t dllname_len = xRtnValue.val.str[0];
	size_t msg_len = leader_len + dllname_len;
	wchar_t *msg_text = new wchar_t[msg_len + 1];
	wcsncpy(msg_text + 1, leader, leader_len);
	wcsncpy(msg_text + 1 + leader_len, xRtnValue.val.str + 1, dllname_len);
	msg_text[0] = msg_len;

// Now the original string has been copied Excel can free it
	Excel12(xlFree, 0, 1, &xRtnValue);
// Now reuse the XLOPER12 for the new string
	xRtnValue.val.str = msg_text;
// Tell Excel to call back into the DLL to free the string
// memory after it has copied out the return value.
	xRtnValue.xltype = xltypeStr | xlbitDLLFree;
	return &xRtnValue;
}

#if 0
void WINAPI xlAutoFree12(LPXLOPER p_oper)
{
	if(p_oper->xltype == (xltypeStr | xlbitDLLFree))
		delete[] p_oper->val.str;
}
#endif

LPXLOPER12 WINAPI get_DLL_name_3(LPXLOPER12 pxArg)
{
	LPXLOPER12 pxRtnValue = new XLOPER12; // Thread-safe

	if(pxArg->xltype != xltypeBool || pxArg->val.xbool != 1)
	{
// If argument not Boolean TRUE return #N/A
		pxRtnValue->xltype = xltypeErr | xlbitDLLFree;
		pxRtnValue->val.err = xlerrNA;
		return pxRtnValue;
	}

	Excel12(xlfGetName, pxRtnValue, 0);

// If xlfGetName failed, pxRtnValue will be #VALUE!
	if(pxRtnValue->xltype != xltypeStr)
	{
// Even though an error type does not point to memory,
// Excel needs to pass this oper to xlAutoFree12 to
// free pxRtnValue itself.
		pxRtnValue->xltype |= xlbitDLLFree;
		return pxRtnValue;
	}

// Make a copy of the DLL path and filename
	wchar_t *leader = L"The full pathname for this DLL is ";
	size_t leader_len = wcslen(leader);
	size_t dllname_len = pxRtnValue->val.str[0];
	size_t msg_len = leader_len + dllname_len;
	wchar_t *msg_text = new wchar_t[msg_len + 1];
	wcsncpy(msg_text + 1, leader, leader_len);
	wcsncpy(msg_text + 1 + leader_len, pxRtnValue->val.str + 1, dllname_len);
	msg_text[0] = msg_len;

// Now the original string has been copied Excel can free it
	Excel12(xlFree, 0, 1, pxRtnValue);
// Now reuse the XLOPER12 for the new string.
	pxRtnValue->val.str = msg_text;
	pxRtnValue->xltype = xltypeStr | xlbitDLLFree;
	return pxRtnValue;
}

#if 0
void WINAPI xlAutoFree12(LPXLOPER p_oper)
{
	if(p_oper->xltype == (xltypeStr | xlbitDLLFree))
		delete[] p_oper->val.str;

	delete p_oper;
}
#endif

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void WINAPI rotate_clockwise_90_FP12(FP12 *pArray)
{
	if(pArray->columns != 2 || pArray->rows != 2)
	{
		memset(pArray->array, 0,
			sizeof(double) * pArray->columns * pArray->rows);
		return; // Returns the original array set to zeros
	}
	double temp[4] = {pArray->array[2], pArray->array[0],
		pArray->array[3], pArray->array[1]};
// Overwrite the passed-in array.  This is OK as we are only
// writing the same number of doubles as the array contained
// when passed in.
	memcpy(pArray->array, temp, 4 * sizeof(double));
}


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#include <stdio.h>

#define MAX_V12_STRBUFFLEN	32678
	
wchar_t * deep_copy_wcs(const wchar_t *p_source)
{
	if(!p_source)
		return NULL;

	size_t source_len = p_source[0];
	bool truncated = false;

	if(source_len + 1 > MAX_V12_STRBUFFLEN)
	{
		source_len = MAX_V12_STRBUFFLEN - 1; // Truncate the copy
		truncated = true;
	}
	wchar_t *p_copy = new wchar_t[source_len + 1];
	wcsncpy(p_copy, p_source, source_len + 1);
	if(truncated)
		p_copy[0] = source_len;
	return p_copy;
}

LPXLOPER12 WINAPI AsText(LPXLOPER12 pArg)
{
	LPXLOPER12 pRtnVal = new XLOPER12;

// Operates on the first element of an array.
// This re-use of the argument is safe.
	if(pArg->xltype == xltypeMulti)
		pArg = pArg->val.array.lparray;

	switch(pArg->xltype)
	{
		case xltypeErr:
			memcpy(pRtnVal, pArg, sizeof(XLOPER12));
			pRtnVal->xltype |= xlbitDLLFree;
			return pRtnVal;

		case xltypeNum:
		case xltypeMissing:
		case xltypeNil:
		case xltypeBool:
			pRtnVal->xltype = xltypeStr | xlbitDLLFree;
			pRtnVal->val.str = new wchar_t[1];
			pRtnVal->val.str[0] = 0;
			return pRtnVal;

		case xltypeStr:
			pRtnVal->xltype = xltypeStr | xlbitDLLFree;
			pRtnVal->val.str = deep_copy_wcs(pArg->val.str);
			return pRtnVal;
		
		default: // xltypeSRef, xltypeRef, xltypeFlow, xltypeInt
			pRtnVal->xltype = xltypeErr | xlbitDLLFree;
			pRtnVal->val.err = xlerrValue;
			return pRtnVal;
	}
}


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

LPXLOPER12 WINAPI mtr_unsafe_example(LPXLOPER12 pxArg)
{
	static XLOPER12 xRetVal; // memory shared by all threads!!!
// code sets xRetVal to a function of pArg ...
	return &xRetVal;
}


LPXLOPER12 WINAPI mtr_safe_example_1(LPXLOPER12 pxArg)
{
// pxRetVal must be freed later by xlAutoFree12
	LPXLOPER12 pxRetVal = new XLOPER12;
// code sets pxRetVal to a function of pxArg ...
	pxRetVal->xltype |= xlbitDLLFree; // Needed for all types
	return pxRetVal; // xlAutoFree12 must free this
}

LPXLOPER get_thread_local_xloper(void);

LPXLOPER WINAPI mtr_safe_example_2(LPXLOPER pxArg)
{
	LPXLOPER pxRetVal = get_thread_local_xloper();
// code sets ret_val to a function of arg setting xlbitDLLFree or
// xlbitXLFree if required
	return pxRetVal; // xlAutoFree must not free this pointer!
}


struct TLS_data
{
	XLOPER xloper_shared_ret_val;
	XLOPER12 xloper12_shared_ret_val;
// Add other required thread-local data here...
};


TLS_data *get_TLS_data(void)
{
// Get a pointer to this thread's static memory
	void *pTLS = TlsGetValue(TlsIndex);
	if(!pTLS) // No TLS memory for this thread yet
	{
		if((pTLS = calloc(1, sizeof(TLS_data))) == NULL)
		// Display some error message (omitted)
			return NULL;
		TlsSetValue(TlsIndex, pTLS); // Associate with this thread
	}
	return (TLS_data *)pTLS;
}


LPXLOPER get_thread_local_xloper(void)
{
	TLS_data *pTLS = get_TLS_data();
	if(pTLS)
		return &(pTLS->xloper_shared_ret_val);
	return NULL;
}

LPXLOPER12 get_thread_local_xloper12(void)
{
	TLS_data *pTLS = get_TLS_data();
	if(pTLS)
		return &(pTLS->xloper12_shared_ret_val);
	return NULL;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#endif // COMPILE_DOC_EXAMPLES

//xltypeStr
#define MAX_XLOPER_STR_LEN		255
#define MAX_XLOPER12_STR_LEN	32767

char *example_static_cs = "Null-terminated byte string. (39 bytes)";
char *example_static_cs2 = "\066Null-terminated length-counted byte string. (54 bytes)";
wchar_t *example_static_ws = L"Null-terminated wide string. (39 chars)";
wchar_t *example_static_ws2 = L"\066Null-terminated length-counted wide string. (54 chars)";


LPXLOPER WINAPI xl_Str_example(int arg)
{
	static XLOPER xRetVal;
	xRetVal.xltype = xltypeStr;
// Return a deep copy of example_static_cs
	size_t len = strlen(example_static_cs);
	if(len > MAX_XLOPER_STR_LEN)
		len = MAX_XLOPER_STR_LEN;

	char *buffer = new char[len + 1]; // +1 for the length char
	memcpy(buffer + 1, example_static_cs, len); // Do not copy null-terminator
	buffer[0] = (BYTE)len; // Cast to unsigned char
	xRetVal.val.str = buffer;
// Buffer must be freed in xlAutoFree
	xRetVal.xltype |= xlbitDLLFree;
	return &xRetVal;
}

LPXLOPER WINAPI xl_Str_example2(int arg)
{
	static XLOPER xRetVal;
// Return a shallow copy of example_static_cs2
// No need to set xlbitDLLFree as the XLOPER
// points to a static string.  No need to make
// a copy or set the length character, as the
// supplied string already has the length
// encoded into the first character.
	xRetVal.xltype = xltypeStr;
	xRetVal.val.str = example_static_cs2;
	return &xRetVal;
}

LPXLOPER12 WINAPI xl12_Str_example(int arg)
{
	static XLOPER12 xRetVal;
	xRetVal.xltype = xltypeStr;
// Return a deep copy of example_static_ws
	size_t len = wcslen(example_static_ws);
	if(len > MAX_XLOPER12_STR_LEN)
		len = MAX_XLOPER12_STR_LEN;

	wchar_t *buffer = new wchar_t[len + 1]; // +1 for the length char
	memcpy((char *)&buffer[1], (const char *)example_static_ws,
		sizeof(wchar_t) * len); // Do not copy null-terminator
	buffer[0] = (wchar_t)len; // Cast to unsigned 16-bit int
	xRetVal.val.str = buffer;
// Buffer must be freed in xlAutoFree12
	xRetVal.xltype |= xlbitDLLFree;
	return &xRetVal;
}

LPXLOPER12 WINAPI xl12_Str_example2(int arg)
{
	static XLOPER12 xRetVal;
// Return a shallow copy of example_static_ws2
// No need to set xlbitDLLFree as the XLOPER12
// points to a static string.  No need to make
// a copy or set the length character, as the
// supplied string already has the length
// encoded into the first character.
	xRetVal.xltype = xltypeStr;
	xRetVal.val.str = example_static_ws2;
	return &xRetVal;
}

//xltypeInt
LPXLOPER WINAPI xl_Int_example(int arg)
{
	static XLOPER xRetVal;
	xRetVal.xltype = xltypeInt;
	xRetVal.val.w = (short)arg; // XLOPER is 16-bit
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

LPXLOPER12 WINAPI xl12_Int_example(int arg)
{
	static XLOPER12 xRetVal;
	xRetVal.xltype = xltypeInt;
	xRetVal.val.w = arg; // XLOPER12 is 32-bit
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

#include <math.h>

//xltypeNum
LPXLOPER WINAPI xl_Num_example(int arg)
{
	static XLOPER xRetVal;
	xRetVal.xltype = xltypeNum;
	xRetVal.val.num = sqrt(fabs(arg));
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

LPXLOPER12 WINAPI xl12_Num_example(int arg)
{
	static XLOPER12 xRetVal;
	xRetVal.xltype = xltypeNum;
	xRetVal.val.num = sqrt(fabs(arg));
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

//xltypeBool
LPXLOPER WINAPI xl_Bool_example(int arg)
{
	static XLOPER xRetVal;
	xRetVal.xltype = xltypeBool;
// Set the XLOPER to TRUE if arg > 0. 1 = TRUE, 0 = FALSE
#ifdef __cplusplus
	xRetVal.val.xbool = (WORD)(arg > 0 ? 1 : 0);
#else
	xRetVal.val.bool = (WORD)(arg > 0 ? 1 : 0);
#endif
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

LPXLOPER12 WINAPI xl12_Bool_example(int arg)
{
	static XLOPER12 xRetVal;
	xRetVal.xltype = xltypeBool;
// Set the XLOPER12 to TRUE if arg > 0. 1 = TRUE, 0 = FALSE
	xRetVal.val.xbool = (arg > 0 ? 1 : 0);
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

//xltypeErr
LPXLOPER WINAPI xl_Err_example(int arg)
{
	static XLOPER xRetVal;
	xRetVal.xltype = xltypeErr;
	if(arg == 0)
		xRetVal.val.w = xlerrDiv0;
	else
		xRetVal.val.w = xlerrNA;
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

LPXLOPER12 WINAPI xl12_Err_example(int arg)
{
	static XLOPER12 xRetVal;
	xRetVal.xltype = xltypeErr;
	if(arg == 0)
		xRetVal.val.w = xlerrDiv0;
	else
		xRetVal.val.w = xlerrNA;
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

//xltypeSRef
LPXLOPER WINAPI xl_SRef_example(int arg)
{
	if(arg < 1)
		arg = 1;
	else if(arg > 100)
		arg = 100;

// Decrement since rows and columns are
// counted from zero.
	arg--;

// Return a reference to a square block
// arg x arg (R1C1:R[arg]C[arg]) on the
// current sheet.
	static XLOPER xRetVal;
	xRetVal.xltype = xltypeSRef;
	xRetVal.val.sref.count = 1; // Always set this to 1
	xRetVal.val.sref.ref.rwFirst = (WORD)0; // Row 1
	xRetVal.val.sref.ref.rwLast = (WORD)arg;
	xRetVal.val.sref.ref.colFirst = (BYTE)0; // Column A
	xRetVal.val.sref.ref.colLast = (BYTE)arg;
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

LPXLOPER12 WINAPI xl_SRef_example12(int arg)
{
	if(arg < 1)
		arg = 1;
	else if(arg > 100)
		arg = 100;

// Decrement since rows and columns are
// counted from zero.
	arg--;

// Return a reference to a square block
// arg x arg (R1C1:R[arg]C[arg]) on the
// current sheet.
	static XLOPER12 xRetVal;
	xRetVal.xltype = xltypeSRef;
	xRetVal.val.sref.count = 1; // Always set this to 1
	xRetVal.val.sref.ref.rwFirst = (RW)0; // Row 1
	xRetVal.val.sref.ref.rwLast = (RW)arg;
	xRetVal.val.sref.ref.colFirst = (COL)0; // Column A
	xRetVal.val.sref.ref.colLast = (COL)arg;
// No need to set xlbitDLLFree, as no memory
// associated with this type
	return &xRetVal;
}

//xltypeRef
LPXLOPER WINAPI xl_Ref_example(int arg)
{
	if(arg < 1)
		arg = 1;
	else if(arg > 100)
		arg = 100;

// Decrement since rows and columns are counted from zero.
	arg--;

// Return a reference to a square block arg x arg
// (R1C1:R[arg]C[arg]) on the specified sheet.
	static XLOPER xRetVal;

// Obtain the internal sheet ID for Sheet1 in Book1
	XLOPER xSheetName;
	xSheetName.xltype = xltypeStr;
	xSheetName.val.str = "\021[BOOK1.XLS]SHEET1";
	if(Excel4(xlSheetId, &xRetVal, 1, &xSheetName) != xlretSuccess
	|| xRetVal.xltype != xltypeRef) // Should always be xltypeRef
	{
// Log or display an error (omitted)
		return NULL;
	}
// xlSheetId returns an xltypeRef with lpmref set to NULL, so
// if we were simply obtaining the sheet ID we would not need
// to free the XLOPER memory using xlFree.

// Allocate memory for an XLMREF structure containing one block
	xRetVal.val.mref.lpmref = new XLMREF[1];
// count is set to one as we are returning a single block
	xRetVal.val.mref.lpmref->count = 1;
	xRetVal.val.mref.lpmref->reftbl[0].rwFirst = (WORD)0;
	xRetVal.val.mref.lpmref->reftbl[0].rwLast = (WORD)arg;
	xRetVal.val.mref.lpmref->reftbl[0].colFirst = (BYTE)0;
	xRetVal.val.mref.lpmref->reftbl[0].colLast = (BYTE)arg;
// Need to set xlbitDLLFree so that xlAutoFree can release lpmref
	xRetVal.xltype |= xlbitDLLFree;
	return &xRetVal;
}


LPXLOPER12 WINAPI xl_Ref_example12(int arg)
{
	if(arg < 1)
		arg = 1;
	else if(arg > 100)
		arg = 100;

// Decrement since rows and columns are counted from zero.
	arg--;

// Return a reference to a square block arg x arg
// (R1C1:R[arg]C[arg]) on the specified sheet.
	static XLOPER12 xRetVal;

// Obtain the internal sheet ID for Sheet1 in Book1
	XLOPER12 xSheetName;
	xSheetName.xltype = xltypeStr;
	xSheetName.val.str = L"\021[BOOK1.XLS]SHEET1";
	if(Excel12(xlSheetId, &xRetVal, 1, &xSheetName) != xlretSuccess
	|| xRetVal.xltype != xltypeRef) // Should always be xltypeRef
	{
// Log or display an error (omitted)
		return NULL;
	}
// xlSheetId returns an xltypeRef with lpmref set to NULL, so
// if we were simply obtaining the sheet ID we would not need
// to free the XLOPER memory using xlFree.

// Allocate memory for an XLMREF structure containing one block
	xRetVal.val.mref.lpmref = new XLMREF12[1];
// count is set to one as we are returning a single block
	xRetVal.val.mref.lpmref->count = 1;
	xRetVal.val.mref.lpmref->reftbl[0].rwFirst = (RW)0;
	xRetVal.val.mref.lpmref->reftbl[0].rwLast = (RW)arg;
	xRetVal.val.mref.lpmref->reftbl[0].colFirst = (COL)0;
	xRetVal.val.mref.lpmref->reftbl[0].colLast = (COL)arg;
// Need to set xlbitDLLFree so that xlAutoFree can release lpmref
	xRetVal.xltype |= xlbitDLLFree;
	return &xRetVal;
}


XLMREF *new_XLMREF(int count)
{
	if(count < 1 || count > 255)
		return NULL;
	size_t size = sizeof(XLMREF) + sizeof(XLREF) * (count - 1);
	XLMREF *pref = (XLMREF *)malloc(size);
	pref->count = count;
	return pref;
}

XLMREF12 *new_XLMREF12(int count)
{
	if(count < 1 || count > 255)
		return NULL;
	size_t size = sizeof(XLMREF12) + sizeof(XLREF12) * (count - 1);
	XLMREF12 *pref = (XLMREF12 *)malloc(size);
	pref->count = count;
	return pref;
}


//xltypeMulti
LPXLOPER WINAPI xl_Multi_example(int arg)
{
	static XLOPER xRetVal;
	xRetVal.xltype = xltypeMulti;
	xRetVal.val.array.rows = (WORD)2;
	xRetVal.val.array.columns = (WORD)2;
	LPXLOPER ptemp = new XLOPER[4];
	xRetVal.val.array.lparray = ptemp;

// 1st element: Row 1, Col 1: Pi = 3.14159265358979...
	ptemp[0].xltype = xltypeNum;
	ptemp[0].val.num = 3.14159265358979;

// 2nd element: Row 1, Col 2: Boolean TRUE
	ptemp[1].xltype = xltypeBool;
#ifdef __cplusplus
	ptemp[1].val.xbool = 1;
#else
	ptemp[1].val.bool = 1;
#endif

// 3rd element: Row 2, Col 1: Error #NA
	ptemp[2].xltype = xltypeErr;
	ptemp[2].val.err = xlerrNA;

// 4th element: Row 2, Col 2: String "Hello world"
// Make a deep copy of the string.
	char *p_msg = "Hello world";
	size_t len = strlen(p_msg);
	ptemp[3].xltype = xltypeStr;
	ptemp[3].val.str = new char[len + 1];
	memcpy(ptemp[3].val.str + 1, p_msg, len);
	ptemp[3].val.str[0] = (BYTE)len;

// Need to set xlbitDLLFree so that xlAutoFree12 can release
// lparray. xlAutoFree12 will also need to release the memory
// for the deep copy of the message "Hello World". This
// code assumes xlAutoFree12 will always free strings within
// DLL-allocated arrays using delete.
	xRetVal.xltype |= xlbitDLLFree;
	return &xRetVal;
}

LPXLOPER12 WINAPI xl12_Multi_example(int arg)
{
	static XLOPER12 xRetVal;
	xRetVal.xltype = xltypeMulti;
	xRetVal.val.array.rows = (RW)2;
	xRetVal.val.array.columns = (COL)2;
	LPXLOPER12 ptemp = new XLOPER12[4];
	xRetVal.val.array.lparray = ptemp;

// 1st element: Row 1, Col 1: Pi = 3.14159265358979...
	ptemp[0].xltype = xltypeNum;
	ptemp[0].val.num = 3.14159265358979;

// 2nd element: Row 1, Col 2: Boolean TRUE
	ptemp[1].xltype = xltypeBool;
	ptemp[1].val.xbool = 1;

// 3rd element: Row 2, Col 1: Error #NA
	ptemp[2].xltype = xltypeErr;
	ptemp[2].val.err = xlerrNA;

// 4th element: Row 2, Col 2: String "Hello world"
// Make a deep copy of the string.
	wchar_t *p_msg = L"Hello world";
	size_t len = wcslen(p_msg);
	ptemp[3].xltype = xltypeStr;
	ptemp[3].val.str = new wchar_t[len + 1];
	memcpy((char *)&(ptemp[3].val.str[1]), (const char *)p_msg, len);
	ptemp[3].val.str[0] = len;

// Need to set xlbitDLLFree so that xlAutoFree can release
// lparray. xlAutoFree will also need to release the memory
// for the deep copy of the message "Hello World". This
// code assumes xlAutoFree will always free strings within
// DLL-allocated arrays using delete.
	xRetVal.xltype |= xlbitDLLFree;
	return &xRetVal;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

int WINAPI IsArgMissing(LPXLOPER pxArg)
{
// Argument registered as type P
	if((pxArg->xltype & (xltypeMissing | xltypeNil)) != 0)
		return 1; // Yes, missing

	return 0; // No, not missing
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

CmdRegData RC_EvaluateName( // Instance name is not important
	"evaluate_name_example",	// Name in code
	"EvaluateName");	// Name that Excel uses

int WINAPI evaluate_name_example(void)
{
	char *expression = "\016!MyDefinedName";
	XLOPER xNameText, xNameValue;
	xNameText.xltype = xltypeStr;
	xNameText.val.str = expression;
// Try to evaluate the name.  Will fail with a #NAME? error
// if MyDefinedName is not defined in the active workbook.
	Excel4(xlfEvaluate, &xNameValue, 1, &xNameText);
// Attempt to convert the value to a string and display it in
// an alert dialog.  This fails if xNameValue is an error value.
	Excel4(xlcAlert, 0, 1, &xNameValue);
// Must free xNameValue in case MyDefinedName evaluated to a string
	Excel4(xlFree, 0, 1, &xNameValue);
	return 1;
}


CmdRegData RC_EvaluateExpr( // Instance name is not important
	"evaluate_epresion_example",	// Name in code
	"EvaluateExpr");	// Name that Excel uses

int WINAPI evaluate_epresion_example(void)
{
	char *expression = "\022COUNT(!A1:IV65536)";
	XLOPER xExprText, xExprValue;
	xExprText.xltype = xltypeStr;
	xExprText.val.str = expression;
// Try to evaluate the formula
	Excel4(xlfEvaluate, &xExprValue, 1, &xExprText);
// Attempt to convert the value to a string and display it in
// an alert dialog.  Will fail if xExprValue is an error
	Excel4(xlcAlert, 0, 1, &xExprValue);
// Not strictly necessary, as COUNT never returns a string
// but does no harm
	Excel4(xlFree, 0, 1, &xExprValue);
	return 1;
}


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#if 0

LPXLOPER TempNum(double d);
LPXLOPER TempStrConst(const LPSTR lpstr);
int cdecl Excel(int xlfn, LPXLOPER pxResult, int count, ...);

double call_ATP_example(void)
{
	XLOPER xPrice;

	int xl_ret_val;

	if(gExcelVersion12plus) // Excel 2007+
	{
		xl_ret_val = Excel(xlfPrice, &xPrice, 7,
			TempNum(39084.0), // settlement date 2-Jan-2007
			TempNum(46706.0), // maturity date 15-Nov-2027
			TempNum(0.04), // Coupon
			TempNum(0.05), // Yield
			TempNum(1.0), // redepmtion value: 100% of face
			TempNum(1.0), // Annual coupons
			TempNum(1.0)); // Rate basis Act/Act
	}
	else // Excel 2003-
	{
		xl_ret_val = Excel4(xlUDF, &xPrice, 8,
			TempStrConst("PRICE"),
			TempNum(39084.0), // settlement date 2-Jan-2007
			TempNum(46706.0), // maturity date 15-Nov-2027
			TempNum(0.04), // Coupon
			TempNum(0.05), // Yield
			TempNum(1.0), // redepmtion value: 100% of face
			TempNum(1.0), // Annual coupons
			TempNum(1.0)); // Rate basis Act/Act
	}
	if(xl_ret_val != xlretSuccess || xPrice.xltype != xltypeNum)
	{
// Even though PRICE is not expected to return a string, there
// is no harm in freeing the XLOPER to be safe
		Excel4(xlFree, 0, 1, &xPrice);
		return -1.0; // an error value
	}
	return xPrice.val.num;
}

#endif

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

FnRegData RF_UDF_1( // Instance name is not important
	"UDF_1",				// Function name as in code (v11-)
	"1P",					// Return type and argument types (v11-)
	"",						// Function name as in code (v12+)
	"",						// Return type and argument types (v12+)
	"UDF_1",				// Fn name as it appears on worksheet
	"Inupt",				// Argument names (Excel 11-: max 30)
	"",						// Argument names (Excel 12+: max 255)
	"Example",				// Function category for Function Wizard
	"",						// Help file (optional)
	"Returns input number + 1",		// Function description text (optional)
	"the inputs  ",			// Fn arg help text strings...
	NULL); // Arg list terminator

// Registered as "1P".  Returns its argument unmodified
// unless it is a number, in which case it increments it
void WINAPI UDF_1(LPXLOPER pxArg)
{
	if(pxArg->xltype == xltypeNum)
		pxArg->val.num += 1.0;
}

FnRegData RF_UDF_2( // Instance name is not important
	"UDF_2",				// Function name as in code (v11-)
	"PP",					// Return type and argument types (v11-)
	"",						// Function name as in code (v12+)
	"",						// Return type and argument types (v12+)
	"UDF_2",				// Fn name as it appears on worksheet
	"Inupt",				// Argument names (Excel 11-: max 30)
	"",						// Argument names (Excel 12+: max 255)
	"Example",				// Function category for Function Wizard
	"",						// Help file (optional)
	"Returns input number + 1",		// Function description text (optional)
	"the inputs  ",			// Fn arg help text strings...
	NULL); // Arg list terminator

// Registered as "PP".  Returns its argument unmodified
// unless it is a number, in which case it increments it.
LPXLOPER WINAPI UDF_2(LPXLOPER pxArg)
{
	static XLOPER xRetVal; // Not thread-safe
	XLOPER xFn;
	xFn.xltype = xltypeStr;
	xFn.val.str = "\005UDF_1";
	Excel4(xlfEvaluate, &xFn, 1, &xFn);
	Excel4(xlUDF, &xRetVal, 2, &xFn, pxArg);
// Need to tell Excel to free memory, as pxArg, and so
// xRetVal, could be a string or an array.
	xRetVal.xltype |= xlbitXLFree;
	return &xRetVal;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#if 0
#define CLASS_NAME_BUFFSIZE    100
#define WINDOW_TEXT_BUFFSIZE   100

// Data structure used as input to xldlg_enum_proc(), called by
// called_from_paste_fn_dlg(), called_from_replace_dlg(), and
// called_from_Excel_dlg().  These functions tell the caller whether
// the current worksheet function was called from one or either of
// these dialogs

typedef struct
{
	bool is_dlg;
	short low_hwnd;
	char *window_title_text; // set to NULL if don't care
}
	xldlg_enum_struct;

// The callback function called by Windows for every top-level window
BOOL __stdcall xldlg_enum_proc(HWND hwnd, xldlg_enum_struct *p_enum)
{
// Check if the parent window is Excel
	if(LOWORD((DWORD)GetParent(hwnd)) != p_enum->low_hwnd)
		return TRUE; // keep iterating

	char class_name[CLASS_NAME_BUFFSIZE + 1];
//	Ensure that class_name is always null terminated for safety
	class_name[CLASS_NAME_BUFFSIZE] = 0;

	GetClassName(hwnd, class_name, CLASS_NAME_BUFFSIZE);

//	Do a case-insensitve comparison for the Excel dialog window
//	class name with the Excel version number truncated
	size_t len; // The length of the window's title text
	if(_strnicmp(class_name, "bosa_sdm_xl", 11) == 0)
	{
// Check if a searching for a specific title string
		if(p_enum->window_title_text) 
		{
// Get the window's title and see if it matches the given text
			char buffer[WINDOW_TEXT_BUFFSIZE + 1];
			buffer[WINDOW_TEXT_BUFFSIZE] = 0;
			len = GetWindowText(hwnd, buffer, WINDOW_TEXT_BUFFSIZE);
			if(len == 0) // No title
			{
				if(p_enum->window_title_text[0] != 0)
					return TRUE; // No match, so keep iterating
			}
// Window has a title so do a case-insensitive comparison of the
// title and the search text, if provided
			else if(p_enum->window_title_text[0] != 0
			&& _stricmp(buffer, p_enum->window_title_text) != 0)
				return TRUE; // Keep iterating
		}
		p_enum->is_dlg = true;
		return FALSE;  // Tells Windows to stop iterating
	}
	return TRUE; // Tells Windows to continue iterating
}

bool called_from_paste_fn_dlg(void)
{
	XLOPER xHwnd;
// Calls Excel4, so only returns the low part of Excel's
// main window handle.  This is OK for the search however.
	if(Excel4(xlGetHwnd, &xHwnd, 0))
		return false; // Couldn't get it, so assume not

// Search for bosa_sdm_xl* dialog with no title string
	xldlg_enum_struct es = {FALSE, xHwnd.val.w, ""};
	EnumWindows((WNDENUMPROC)xldlg_enum_proc, (LPARAM)&es);
	return es.is_dlg;
}


// Modified MS code
//
// Structure to pass the the EnumWindowsProc
//
typedef struct _EnumStruct
{
	HWND        hwnd;	// Return value for Excel's main hWnd
	unsigned short  wLoword; //Contains LowWord of Excel's main hWnd
} EnumStruct;

#define CLASS_NAME_BUFFER	50

BOOL CALLBACK EnumProc(HWND hwnd, EnumStruct * pEnum)
{
	// first check the class of the window.  Must be "XLMAIN"
	char rgsz[CLASS_NAME_BUFFER];

	// check the loword of the window handle
	if (LOWORD((DWORD) hwnd) == pEnum->wLoword)
	{
	// if that hits, check that the class name is XLMAIN
		GetClassName(hwnd, rgsz, CLASS_NAME_BUFFER);
		if (!lstrcmpi(rgsz, "XLMAIN"))
		{
			// We have a match, return Excel's main hWnd
			pEnum->hwnd = hwnd;
			return FALSE;
		}
	}

	// no luck - continue the enumeration
	return TRUE;
}


#endif


