//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file (and header file) contain the definitions of various
// static structures and strings.  The structures contain the data passed to
// Excel when registering commands and functions.
//============================================================================
//============================================================================
#include <windows.h>

#include "XllAddIn.h"

char *AddinVersionStr = "Skeleton Excel add-in V1";
char *AddinName = "Skeleton";

//---------------------------------------------------------
//---------------------------------------------------------
// Exported command information
//---------------------------------------------------------
//---------------------------------------------------------
CmdRegData RC_ExampleFn( // Instance name is not important
	"ExampleCmdExport",	// Name in code
	"ExampleCmdExport");	// Name that Excel uses

int __stdcall ExampleCmdExport(void)
{
	cpp_xloper Op("ExampleCmdExport command");
	Op.Alert();
	return 1;
}

//---------------------------------------------------------
//---------------------------------------------------------
// Exported worksheet function information
//---------------------------------------------------------
//---------------------------------------------------------
FnRegData RF_ExampleFn( // Instance name is not important
	"ExampleFnExport_xl4",	// Function name as in code (v11-)
	"RP",					// Return type and argument types (v11-)
	"ExampleFnExport_xl12",	// Function name as in code (v12+)
	"UQ$",					// Return type and argument types (v12+)
	"ExampleFnExport",		// Fn name as it appears on worksheet
	"Inputs",				// Argument names (Excel 11-: max 30)
	"",						// Argument names (Excel 12+: max 255)
	"Example",				// Function category for Function Wizard
	"",						// Help file (optional)
	"Returns something",	// Function description text (optional)
	"the inputs  ",			// Fn arg help text strings...
	NULL); // Arg list terminator


//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers
xloper * __stdcall ExampleFnExport_xl4(xloper *inputs)
{
#pragma EXPORT
	cpp_xloper RetVal("ExampleFnExport_xl4 function");
	return RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s
xloper12 * __stdcall ExampleFnExport_xl12(xloper12 *inputs)
{
#pragma EXPORT
	cpp_xloper RetVal("ExampleFnExport_xl12 function");
	return RetVal.ExtractXloper12();
}

//============================================================================
// Can only call xlfGetWorkspace from a macro-sheet function or command, so if
// that fails, then is a worksheet function
bool caller_is_ws_function(void)
{
// Retrieve the R1C1 state. Fails if called from WS function
// Don't need the return value so pass NULL.
	if(gExcelVersion12plus)
	{
		xloper12 four;
		four.val.w = 4; // Ask if Excel is in R1C1 mode
		four.xltype = xltypeInt;
		return Excel12(xlfGetWorkspace, 0, 1, &four) != xlretSuccess;
	}
	else
	{
		xloper four;
		four.val.w = 4; // Ask if Excel is in R1C1 mode
		four.xltype = xltypeInt;
		return Excel4(xlfGetWorkspace, 0, 1, &four) != xlretSuccess;
	}
}
