//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================

//============================================================================
//============================================================================
// This file contains definitions of some (but not necessarily all) of the
// XLL's exported function interfaces.  These function interfaces in general
// call into core code in other modules or libraries in this project.
//============================================================================
//============================================================================
#include <windows.h>

#include "XllAddIn.h"
#include "background.h"

char *AddinVersionStr = "Skeleton Excel add-in V1";
char *AddinName = "Skeleton";

//---------------------------------------------------------
//---------------------------------------------------------
// Exported command information
//---------------------------------------------------------
//---------------------------------------------------------
CmdRegData RC_ExampleFn( // Instance name is not important
	"ExampleCmdExport",	// Name in code
	"ExampleCmdExport");	// Name that Excel uses

int __stdcall ExampleCmdExport(void)
{
	cpp_xloper Op("ExampleCmdExport command");
	Op.Alert();
	Op.Excel(xlfCaller);
	if(Op.IsMulti())
	{
		cpp_xloper Elt;
		DWORD size;
		if(Op.GetArraySize(size) && size > 0)
			for(DWORD i = 0; i < size; i++)
			{
				Op.GetArrayElt(i, Elt);
				Elt.Alert();
			}
	}

	return 1;
}

//---------------------------------------------------------
//---------------------------------------------------------
// Exported worksheet function information
//---------------------------------------------------------
//---------------------------------------------------------
FnRegData RF_ExampleFn( // Instance name is not important
	"ExampleFnExport_xl4",	// Function name as in code (v11-)
	"RP",					// Return type and argument types (v11-)
	"ExampleFnExport_xl12",	// Function name as in code (v12+)
	"UQ$",					// Return type and argument types (v12+)
	"ExampleFnExport",		// Fn name as it appears on worksheet
	"Inputs",				// Argument names (Excel 11-: max 30)
	"",						// Argument names (Excel 12+: max 255)
	"Example",				// Function category for Function Wizard
	"",						// Help file (optional)
	"Returns something",	// Function description text (optional)
	"the inputs  ",			// Fn arg help text strings...
	NULL); // Arg list terminator


//----------------------------------------------------------------------------
// Excel 11- interface function.  Uses xlopers
xloper * __stdcall ExampleFnExport_xl4(xloper *inputs)
{
#pragma EXPORT
	cpp_xloper RetVal("ExampleFnExport_xl4 function");
	return RetVal.ExtractXloper();
}
//----------------------------------------------------------------------------
// Excel 12+ interface function.  Uses xloper12s
xloper12 * __stdcall ExampleFnExport_xl12(xloper12 *inputs)
{
#pragma EXPORT
	cpp_xloper RetVal("ExampleFnExport_xl12 function");
	return RetVal.ExtractXloper12();
}

//============================================================================
// long_task_interface() is a worksheet function called
// directly by Excel from the foreground thread.  It is only
// required to check arguments and call ExampleTaskList.UpdateTask()
// which returns either an error, or the intermediate or final value
// of the calculation.  UpdateTask() errors can be returned directly
// or, as in this case, the function can return the current
// (previous) value of the calling cell.  This function is registered
// with Excel as a volatile macro sheet function.  As a macro sheet
// function it will only be recalculated by Excel 2007 on the main thread.
//============================================================================
bool LongTaskExampleMain(task *pTask);

FnRegData RF_LongTaskExampleFn( // Instance name is not important
	"LongTaskExampleFn",	// Function name as in code (v11-)
	"RP#!",					// Return type and argument types (v11-)
	"",						// Function name as in code (v12+)
	"",						// Return type and argument types (v12+)
	"LongTaskExampleFn",	// Fn name as it appears on worksheet
	"Inputs",				// Argument names (Excel 11-: max 30)
	"",						// Argument names (Excel 12+: max 255)
	"Example",				// Function category for Function Wizard
	"",						// Help file (optional)
	"Do something in background",	// Function description text (optional)
	"the inputs  ",			// Fn arg help text strings...
	NULL); // Arg list terminator

xloper * __stdcall LongTaskExampleFn(xloper *arg)
{
#pragma EXPORT

	if(called_from_Excel_dlg())
		return p_xlErrNa;

	if(arg->xltype != xltypeNum || arg->val.num < 1)
		return p_xlErrValue;

	cpp_xloper InArray[1] = {arg}; // only 1 argument in this case
	cpp_xloper RetVal;
// UpdateTask makes deep copies of all the supplied arguments
	ExampleTaskList.UpdateTask(LongTaskExampleMain, InArray, 1, RetVal);

	if(RetVal.IsErr())
	{
		WORD err_val;
		RetVal.GetErrVal(err_val);

		switch(err_val)
		{
		// the arguments were not valid
		case xlerrValue:
			break;

		// task has never been completed and is now pending or current
		case xlerrNum:
			break;

		// the thread is inactive
		case xlerrNA:
			break;
		}
// Set RetVal to the existing cell value.  Need macro-sheet
// permissions to do this, and so this line is not thread-safe.
// under Excel 2007.  If thread-safety is required, return
// some other value.
		RetVal.SetToCallerValue();
	}
	return RetVal.ExtractXloper();
}

